# Copyright (C) 2019-2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import collections
import copy
import datetime
import hashlib
from typing import Any, List, Optional, Tuple, Union

import attr
from attrs_strict import AttributeTypeError
import dateutil
from hypothesis import given
from hypothesis.strategies import binary, none
import pytest

from swh.model.collections import ImmutableDict
from swh.model.from_disk import DentryPerms
import swh.model.git_objects
from swh.model.hashutil import MultiHash, hash_to_bytes
import swh.model.hypothesis_strategies as strategies
import swh.model.model
from swh.model.model import (
    BaseModel,
    Content,
    Directory,
    DirectoryEntry,
    MetadataAuthority,
    MetadataAuthorityType,
    MetadataFetcher,
    MissingData,
    Origin,
    OriginVisit,
    OriginVisitStatus,
    Person,
    RawExtrinsicMetadata,
    Release,
    Revision,
    SkippedContent,
    Snapshot,
    TargetType,
    Timestamp,
    TimestampWithTimezone,
    optimized_validator,
)
import swh.model.swhids
from swh.model.swhids import CoreSWHID, ExtendedSWHID, ObjectType
from swh.model.tests.swh_model_data import TEST_OBJECTS
from swh.model.tests.test_identifiers import (
    TS_DATETIMES,
    TS_TIMEZONES,
    directory_example,
    metadata_example,
    release_example,
    revision_example,
    snapshot_example,
)

EXAMPLE_HASH = hash_to_bytes("94a9ed024d3859793618152ea559a168bbcbb5e2")


@given(strategies.objects())
def test_todict_inverse_fromdict(objtype_and_obj):
    (obj_type, obj) = objtype_and_obj

    if obj_type in ("origin", "origin_visit"):
        return

    obj_as_dict = obj.to_dict()
    obj_as_dict_copy = copy.deepcopy(obj_as_dict)

    # Check the composition of to_dict and from_dict is the identity
    assert obj == type(obj).from_dict(obj_as_dict)

    # Check from_dict() does not change the input dict
    assert obj_as_dict == obj_as_dict_copy

    # Check the composition of from_dict and to_dict is the identity
    assert obj_as_dict == type(obj).from_dict(obj_as_dict).to_dict()


@given(strategies.objects())
def test_repr(objtype_and_obj):
    """Checks every model object has a working repr(), and that it can be eval()uated
    (so that printed objects can be copy-pasted to write test cases.)"""
    (obj_type, obj) = objtype_and_obj

    r = repr(obj)
    env = {
        "tzutc": lambda: datetime.timezone.utc,
        "tzfile": dateutil.tz.tzfile,
        "hash_to_bytes": hash_to_bytes,
        **swh.model.swhids.__dict__,
        **swh.model.model.__dict__,
    }
    assert eval(r, env) == obj


@attr.s
class Cls1:
    pass


@attr.s
class Cls2(Cls1):
    pass


_custom_namedtuple = collections.namedtuple("_custom_namedtuple", "a b")


class _custom_tuple(tuple):
    pass


# List of (type, valid_values, invalid_values)
_TYPE_VALIDATOR_PARAMETERS: List[Tuple[Any, List[Any], List[Any]]] = [
    # base types:
    (
        bool,
        [True, False],
        [-1, 0, 1, 42, 1000, None, "123", 0.0, (), ("foo",), ImmutableDict()],
    ),
    (
        int,
        [-1, 0, 1, 42, 1000, DentryPerms.directory, True, False],
        [None, "123", 0.0, (), ImmutableDict()],
    ),
    (
        float,
        [-1.0, 0.0, 1.0, float("infinity"), float("NaN")],
        [True, False, None, 1, "1.2", (), ImmutableDict()],
    ),
    (
        bytes,
        [b"", b"123"],
        [None, bytearray(b"\x12\x34"), "123", 0, 123, (), (1, 2, 3), ImmutableDict()],
    ),
    (str, ["", "123"], [None, b"123", b"", 0, (), (1, 2, 3), ImmutableDict()]),
    (None, [None], [b"", b"123", "", "foo", 0, 123, ImmutableDict(), float("NaN")]),
    # unions:
    (
        Optional[int],
        [None, -1, 0, 1, 42, 1000, DentryPerms.directory],
        ["123", 0.0, (), ImmutableDict()],
    ),
    (
        Optional[bytes],
        [None, b"", b"123"],
        ["123", "", 0, (), (1, 2, 3), ImmutableDict()],
    ),
    (
        Union[str, bytes],
        ["", "123", b"123", b""],
        [None, 0, (), (1, 2, 3), ImmutableDict()],
    ),
    (
        Union[str, bytes, None],
        ["", "123", b"123", b"", None],
        [0, (), (1, 2, 3), ImmutableDict()],
    ),
    # tuples
    (
        Tuple[str, str],
        [("foo", "bar"), ("", ""), _custom_namedtuple("", ""), _custom_tuple(("", ""))],
        [("foo",), ("foo", "bar", "baz"), ("foo", 42), (42, "foo")],
    ),
    (
        Tuple[bytes, bytes],
        [
            (b"foo", b"bar"),
            (b"", b""),
            _custom_namedtuple(b"", b""),
            _custom_tuple((b"", b"")),
        ],
        [(b"foo",), (b"foo", b"bar", b"baz"), (b"foo", 42), (42, b"foo")],
    ),
    (
        Tuple[str, ...],
        [
            ("foo",),
            ("foo", "bar"),
            ("", ""),
            ("foo", "bar", "baz"),
            _custom_namedtuple("", ""),
            _custom_tuple(("", "")),
        ],
        [("foo", 42), (42, "foo")],
    ),
    # composite generic:
    (
        Tuple[Union[str, int], Union[str, int]],
        [("foo", "foo"), ("foo", 42), (42, "foo"), (42, 42)],
        [("foo", b"bar"), (b"bar", "foo")],
    ),
    (
        Union[Tuple[str, str], Tuple[int, int]],
        [("foo", "foo"), (42, 42)],
        [("foo", b"bar"), (b"bar", "foo"), ("foo", 42), (42, "foo")],
    ),
    (
        Tuple[Tuple[bytes, bytes], ...],
        [(), ((b"foo", b"bar"),), ((b"foo", b"bar"), (b"baz", b"qux"))],
        [((b"foo", "bar"),), ((b"foo", b"bar"), ("baz", b"qux"))],
    ),
    # standard types:
    (
        datetime.datetime,
        [
            datetime.datetime(2021, 12, 15, 12, 59, 27),
            datetime.datetime(2021, 12, 15, 12, 59, 27, tzinfo=datetime.timezone.utc),
        ],
        [None, 123],
    ),
    # ImmutableDict
    (
        ImmutableDict[str, int],
        [
            ImmutableDict(),
            ImmutableDict({"foo": 42}),
            ImmutableDict({"foo": 42, "bar": 123}),
        ],
        [ImmutableDict({"foo": "bar"}), ImmutableDict({42: 123})],
    ),
    # Any:
    (
        object,
        [-1, 0, 1, 42, 1000, None, "123", 0.0, (), ImmutableDict()],
        [],
    ),
    (
        Any,
        [-1, 0, 1, 42, 1000, None, "123", 0.0, (), ImmutableDict()],
        [],
    ),
    (
        ImmutableDict[Any, int],
        [
            ImmutableDict(),
            ImmutableDict({"foo": 42}),
            ImmutableDict({"foo": 42, "bar": 123}),
            ImmutableDict({42: 123}),
        ],
        [ImmutableDict({"foo": "bar"})],
    ),
    (
        ImmutableDict[str, Any],
        [
            ImmutableDict(),
            ImmutableDict({"foo": 42}),
            ImmutableDict({"foo": "bar"}),
            ImmutableDict({"foo": 42, "bar": 123}),
        ],
        [ImmutableDict({42: 123})],
    ),
    # attr objects:
    (
        Timestamp,
        [
            Timestamp(seconds=123, microseconds=0),
        ],
        [None, "2021-09-28T11:27:59", 123],
    ),
    (
        Cls1,
        [Cls1(), Cls2()],
        [None, b"abcd"],
    ),
    # enums:
    (
        TargetType,
        [TargetType.CONTENT, TargetType.ALIAS],
        ["content", "alias", 123, None],
    ),
]


@pytest.mark.parametrize(
    "type_,value",
    [
        pytest.param(type_, value, id=f"type={type_}, value={value}")
        for (type_, values, _) in _TYPE_VALIDATOR_PARAMETERS
        for value in values
    ],
)
def test_optimized_type_validator_valid(type_, value):
    validator = optimized_validator(type_)
    validator(None, attr.ib(type=type_), value)


@pytest.mark.parametrize(
    "type_,value",
    [
        pytest.param(type_, value, id=f"type={type_}, value={value}")
        for (type_, _, values) in _TYPE_VALIDATOR_PARAMETERS
        for value in values
    ],
)
def test_optimized_type_validator_invalid(type_, value):
    validator = optimized_validator(type_)
    with pytest.raises(AttributeTypeError):
        validator(None, attr.ib(type=type_), value)


@pytest.mark.parametrize("object_type, objects", TEST_OBJECTS.items())
def test_swh_model_todict_fromdict(object_type, objects):
    """checks model objects in swh_model_data are in correct shape"""
    assert objects
    for obj in objects:
        # Check the composition of from_dict and to_dict is the identity
        obj_as_dict = obj.to_dict()
        assert obj == type(obj).from_dict(obj_as_dict)
        assert obj_as_dict == type(obj).from_dict(obj_as_dict).to_dict()


def test_unique_key():
    url = "http://example.org/"
    date = datetime.datetime.now(tz=datetime.timezone.utc)
    id_ = b"42" * 10
    assert Origin(url=url).unique_key() == {"url": url}
    assert OriginVisit(origin=url, date=date, type="git").unique_key() == {
        "origin": url,
        "date": str(date),
    }
    assert OriginVisitStatus(
        origin=url, visit=42, date=date, status="created", snapshot=None
    ).unique_key() == {
        "origin": url,
        "visit": "42",
        "date": str(date),
    }

    assert Snapshot.from_dict({**snapshot_example, "id": id_}).unique_key() == id_
    assert Release.from_dict({**release_example, "id": id_}).unique_key() == id_
    assert Revision.from_dict({**revision_example, "id": id_}).unique_key() == id_
    assert Directory.from_dict({**directory_example, "id": id_}).unique_key() == id_
    assert (
        RawExtrinsicMetadata.from_dict({**metadata_example, "id": id_}).unique_key()
        == id_
    )

    cont = Content.from_data(b"foo")
    assert cont.unique_key().hex() == "0beec7b5ea3f0fdbc95d0dd47f3c5bc275da8a33"

    kwargs = {
        **cont.to_dict(),
        "reason": "foo",
        "status": "absent",
    }
    del kwargs["data"]
    assert SkippedContent(**kwargs).unique_key() == cont.hashes()


# Anonymization


@given(strategies.objects())
def test_anonymization(objtype_and_obj):
    (obj_type, obj) = objtype_and_obj

    def check_person(p):
        if p is not None:
            assert p.name is None
            assert p.email is None
            assert len(p.fullname) == 32

    anon_obj = obj.anonymize()
    if obj_type == "person":
        assert anon_obj is not None
        check_person(anon_obj)
    elif obj_type == "release":
        assert anon_obj is not None
        check_person(anon_obj.author)
    elif obj_type == "revision":
        assert anon_obj is not None
        check_person(anon_obj.author)
        check_person(anon_obj.committer)
    else:
        assert anon_obj is None


# Origin, OriginVisit, OriginVisitStatus


@given(strategies.origins())
def test_todict_origins(origin):
    obj = origin.to_dict()

    assert "type" not in obj
    assert type(origin)(url=origin.url) == type(origin).from_dict(obj)


@given(strategies.origin_visits())
def test_todict_origin_visits(origin_visit):
    obj = origin_visit.to_dict()

    assert origin_visit == type(origin_visit).from_dict(obj)


def test_origin_visit_naive_datetime():
    with pytest.raises(ValueError, match="must be a timezone-aware datetime"):
        OriginVisit(
            origin="http://foo/",
            date=datetime.datetime.now(),
            type="git",
        )


@given(strategies.origin_visit_statuses())
def test_todict_origin_visit_statuses(origin_visit_status):
    obj = origin_visit_status.to_dict()

    assert origin_visit_status == type(origin_visit_status).from_dict(obj)


def test_origin_visit_status_naive_datetime():
    with pytest.raises(ValueError, match="must be a timezone-aware datetime"):
        OriginVisitStatus(
            origin="http://foo/",
            visit=42,
            date=datetime.datetime.now(),
            status="ongoing",
            snapshot=None,
        )


@pytest.fixture
def origin_visit_status_example():
    tz = datetime.timezone(datetime.timedelta(minutes=+60))
    return OriginVisitStatus(
        origin="http://foo/",
        visit=42,
        date=datetime.datetime.now(tz=tz),
        status="full",
        snapshot=hash_to_bytes("6e65b86363953b780d92b0a928f3e8fcdd10db36"),
    )


def test_origin_visit_status_snapshot_swhid(origin_visit_status_example):
    assert origin_visit_status_example.snapshot_swhid() == CoreSWHID.from_string(
        "swh:1:snp:6e65b86363953b780d92b0a928f3e8fcdd10db36"
    )


def test_origin_visit_status_origin_swhid(origin_visit_status_example):
    assert origin_visit_status_example.origin_swhid() == ExtendedSWHID.from_string(
        "swh:1:ori:e0cee4b024ab93b037a1c182865942f5430c6fa4"
    )


# Timestamp


@given(strategies.timestamps())
def test_timestamps_strategy(timestamp):
    attr.validate(timestamp)


def test_timestamp_seconds():
    attr.validate(Timestamp(seconds=0, microseconds=0))
    with pytest.raises(AttributeTypeError):
        Timestamp(seconds="0", microseconds=0)

    attr.validate(Timestamp(seconds=2**63 - 1, microseconds=0))
    with pytest.raises(ValueError):
        Timestamp(seconds=2**63, microseconds=0)

    attr.validate(Timestamp(seconds=-(2**63), microseconds=0))
    with pytest.raises(ValueError):
        Timestamp(seconds=-(2**63) - 1, microseconds=0)


def test_timestamp_microseconds():
    attr.validate(Timestamp(seconds=0, microseconds=0))
    with pytest.raises(AttributeTypeError):
        Timestamp(seconds=0, microseconds="0")

    attr.validate(Timestamp(seconds=0, microseconds=10**6 - 1))
    with pytest.raises(ValueError):
        Timestamp(seconds=0, microseconds=10**6)

    with pytest.raises(ValueError):
        Timestamp(seconds=0, microseconds=-1)


def test_timestamp_from_dict():
    assert Timestamp.from_dict({"seconds": 10, "microseconds": 5})

    with pytest.raises(AttributeTypeError):
        Timestamp.from_dict({"seconds": "10", "microseconds": 5})

    with pytest.raises(AttributeTypeError):
        Timestamp.from_dict({"seconds": 10, "microseconds": "5"})
    with pytest.raises(ValueError):
        Timestamp.from_dict({"seconds": 0, "microseconds": -1})

    Timestamp.from_dict({"seconds": 0, "microseconds": 10**6 - 1})
    with pytest.raises(ValueError):
        Timestamp.from_dict({"seconds": 0, "microseconds": 10**6})


# TimestampWithTimezone


def test_timestampwithtimezone():
    ts = Timestamp(seconds=0, microseconds=0)
    tstz = TimestampWithTimezone(timestamp=ts, offset_bytes=b"+0000")
    attr.validate(tstz)
    assert tstz.offset_minutes() == 0
    assert tstz.offset_bytes == b"+0000"

    tstz = TimestampWithTimezone(timestamp=ts, offset_bytes=b"+0010")
    attr.validate(tstz)
    assert tstz.offset_minutes() == 10
    assert tstz.offset_bytes == b"+0010"

    tstz = TimestampWithTimezone(timestamp=ts, offset_bytes=b"-0010")
    attr.validate(tstz)
    assert tstz.offset_minutes() == -10
    assert tstz.offset_bytes == b"-0010"

    tstz = TimestampWithTimezone(timestamp=ts, offset_bytes=b"-0000")
    attr.validate(tstz)
    assert tstz.offset_minutes() == 0
    assert tstz.offset_bytes == b"-0000"

    tstz = TimestampWithTimezone(timestamp=ts, offset_bytes=b"-1030")
    attr.validate(tstz)
    assert tstz.offset_minutes() == -630
    assert tstz.offset_bytes == b"-1030"

    tstz = TimestampWithTimezone(timestamp=ts, offset_bytes=b"+1320")
    attr.validate(tstz)
    assert tstz.offset_minutes() == 800
    assert tstz.offset_bytes == b"+1320"

    tstz = TimestampWithTimezone(timestamp=ts, offset_bytes=b"+200")
    attr.validate(tstz)
    assert tstz.offset_minutes() == 120
    assert tstz.offset_bytes == b"+200"

    tstz = TimestampWithTimezone(timestamp=ts, offset_bytes=b"+02")
    attr.validate(tstz)
    assert tstz.offset_minutes() == 120
    assert tstz.offset_bytes == b"+02"

    tstz = TimestampWithTimezone(timestamp=ts, offset_bytes=b"+2000000000")
    attr.validate(tstz)
    assert tstz.offset_minutes() == 0
    assert tstz.offset_bytes == b"+2000000000"

    with pytest.raises(AttributeTypeError):
        TimestampWithTimezone(timestamp=datetime.datetime.now(), offset_bytes=b"+0000")

    with pytest.raises((AttributeTypeError, TypeError)):
        TimestampWithTimezone(timestamp=ts, offset_bytes=0)


def test_timestampwithtimezone_from_datetime():
    # Typical case
    tz = datetime.timezone(datetime.timedelta(minutes=+60))
    date = datetime.datetime(2020, 2, 27, 14, 39, 19, tzinfo=tz)
    tstz = TimestampWithTimezone.from_datetime(date)
    assert tstz == TimestampWithTimezone(
        timestamp=Timestamp(
            seconds=1582810759,
            microseconds=0,
        ),
        offset_bytes=b"+0100",
    )

    # Typical case (close to epoch)
    tz = datetime.timezone(datetime.timedelta(minutes=+60))
    date = datetime.datetime(1970, 1, 1, 1, 0, 5, tzinfo=tz)
    tstz = TimestampWithTimezone.from_datetime(date)
    assert tstz == TimestampWithTimezone(
        timestamp=Timestamp(
            seconds=5,
            microseconds=0,
        ),
        offset_bytes=b"+0100",
    )

    # non-integer number of seconds before UNIX epoch
    date = datetime.datetime(
        1969, 12, 31, 23, 59, 59, 100000, tzinfo=datetime.timezone.utc
    )
    tstz = TimestampWithTimezone.from_datetime(date)
    assert tstz == TimestampWithTimezone(
        timestamp=Timestamp(
            seconds=-1,
            microseconds=100000,
        ),
        offset_bytes=b"+0000",
    )

    # non-integer number of seconds in both the timestamp and the offset
    tz = datetime.timezone(datetime.timedelta(microseconds=-600000))
    date = datetime.datetime(1969, 12, 31, 23, 59, 59, 600000, tzinfo=tz)
    tstz = TimestampWithTimezone.from_datetime(date)
    assert tstz == TimestampWithTimezone(
        timestamp=Timestamp(
            seconds=0,
            microseconds=200000,
        ),
        offset_bytes=b"+0000",
    )

    # timezone offset with non-integer number of seconds, for dates before epoch
    # we round down to the previous second, so it should be the same as
    # 1969-01-01T23:59:59Z
    tz = datetime.timezone(datetime.timedelta(microseconds=900000))
    date = datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=tz)
    tstz = TimestampWithTimezone.from_datetime(date)
    assert tstz == TimestampWithTimezone(
        timestamp=Timestamp(
            seconds=-1,
            microseconds=100000,
        ),
        offset_bytes=b"+0000",
    )


def test_timestampwithtimezone_from_naive_datetime():
    date = datetime.datetime(2020, 2, 27, 14, 39, 19)

    with pytest.raises(ValueError, match="datetime without timezone"):
        TimestampWithTimezone.from_datetime(date)


def test_timestampwithtimezone_from_iso8601():
    date = "2020-02-27 14:39:19.123456+0100"

    tstz = TimestampWithTimezone.from_iso8601(date)

    assert tstz == TimestampWithTimezone(
        timestamp=Timestamp(
            seconds=1582810759,
            microseconds=123456,
        ),
        offset_bytes=b"+0100",
    )


def test_timestampwithtimezone_from_iso8601_negative_utc():
    date = "2020-02-27 13:39:19-0000"

    tstz = TimestampWithTimezone.from_iso8601(date)

    assert tstz == TimestampWithTimezone(
        timestamp=Timestamp(
            seconds=1582810759,
            microseconds=0,
        ),
        offset_bytes=b"-0000",
    )


@pytest.mark.parametrize("date", TS_DATETIMES)
@pytest.mark.parametrize("tz", TS_TIMEZONES)
@pytest.mark.parametrize("microsecond", [0, 1, 10, 100, 1000, 999999])
def test_timestampwithtimezone_to_datetime(date, tz, microsecond):
    date = date.replace(tzinfo=tz, microsecond=microsecond)
    tstz = TimestampWithTimezone.from_datetime(date)

    assert tstz.to_datetime() == date
    assert tstz.to_datetime().utcoffset() == date.utcoffset()


def test_person_from_fullname():
    """The author should have name, email and fullname filled."""
    actual_person = Person.from_fullname(b"tony <ynot@dagobah>")
    assert actual_person == Person(
        fullname=b"tony <ynot@dagobah>",
        name=b"tony",
        email=b"ynot@dagobah",
    )


def test_person_from_fullname_no_email():
    """The author and fullname should be the same as the input (author)."""
    actual_person = Person.from_fullname(b"tony")
    assert actual_person == Person(
        fullname=b"tony",
        name=b"tony",
        email=None,
    )


def test_person_from_fullname_empty_person():
    """Empty person has only its fullname filled with the empty
    byte-string.

    """
    actual_person = Person.from_fullname(b"")
    assert actual_person == Person(
        fullname=b"",
        name=None,
        email=None,
    )


def test_git_author_line_to_author():
    # edge case out of the way
    with pytest.raises(TypeError):
        Person.from_fullname(None)

    tests = {
        b"a <b@c.com>": Person(
            name=b"a",
            email=b"b@c.com",
            fullname=b"a <b@c.com>",
        ),
        b"<foo@bar.com>": Person(
            name=None,
            email=b"foo@bar.com",
            fullname=b"<foo@bar.com>",
        ),
        b"malformed <email": Person(
            name=b"malformed", email=b"email", fullname=b"malformed <email"
        ),
        b'malformed <"<br"@ckets>': Person(
            name=b"malformed",
            email=b'"<br"@ckets',
            fullname=b'malformed <"<br"@ckets>',
        ),
        b"trailing <sp@c.e> ": Person(
            name=b"trailing",
            email=b"sp@c.e",
            fullname=b"trailing <sp@c.e> ",
        ),
        b"no<sp@c.e>": Person(
            name=b"no",
            email=b"sp@c.e",
            fullname=b"no<sp@c.e>",
        ),
        b" more   <sp@c.es>": Person(
            name=b"more",
            email=b"sp@c.es",
            fullname=b" more   <sp@c.es>",
        ),
        b" <>": Person(
            name=None,
            email=None,
            fullname=b" <>",
        ),
    }

    for person in sorted(tests):
        expected_person = tests[person]
        assert expected_person == Person.from_fullname(person)


def test_person_comparison():
    """Check only the fullname attribute is used to compare Person objects"""
    person = Person(fullname=b"p1", name=None, email=None)
    assert attr.evolve(person, name=b"toto") == person
    assert attr.evolve(person, email=b"toto@example.com") == person

    person = Person(fullname=b"", name=b"toto", email=b"toto@example.com")
    assert attr.evolve(person, fullname=b"dude") != person


# Content


def test_content_get_hash():
    hashes = dict(sha1=b"foo", sha1_git=b"bar", sha256=b"baz", blake2s256=b"qux")
    c = Content(length=42, status="visible", **hashes)
    for (hash_name, hash_) in hashes.items():
        assert c.get_hash(hash_name) == hash_


def test_content_hashes():
    hashes = dict(sha1=b"foo", sha1_git=b"bar", sha256=b"baz", blake2s256=b"qux")
    c = Content(length=42, status="visible", **hashes)
    assert c.hashes() == hashes


def test_content_data():
    c = Content(
        length=42,
        status="visible",
        data=b"foo",
        sha1=b"foo",
        sha1_git=b"bar",
        sha256=b"baz",
        blake2s256=b"qux",
    )
    assert c.with_data() == c


def test_content_data_missing():
    c = Content(
        length=42,
        status="visible",
        sha1=b"foo",
        sha1_git=b"bar",
        sha256=b"baz",
        blake2s256=b"qux",
    )
    with pytest.raises(MissingData):
        c.with_data()


@given(strategies.present_contents_d())
def test_content_from_dict(content_d):
    c = Content.from_data(**content_d)
    assert c
    assert c.ctime == content_d["ctime"]

    content_d2 = c.to_dict()
    c2 = Content.from_dict(content_d2)
    assert c2.ctime == c.ctime


def test_content_from_dict_str_ctime():
    # test with ctime as a string
    n = datetime.datetime(2020, 5, 6, 12, 34, tzinfo=datetime.timezone.utc)
    content_d = {
        "ctime": n.isoformat(),
        "data": b"",
        "length": 0,
        "sha1": b"\x00",
        "sha256": b"\x00",
        "sha1_git": b"\x00",
        "blake2s256": b"\x00",
    }
    c = Content.from_dict(content_d)
    assert c.ctime == n


def test_content_from_dict_str_naive_ctime():
    # test with ctime as a string
    n = datetime.datetime(2020, 5, 6, 12, 34)
    content_d = {
        "ctime": n.isoformat(),
        "data": b"",
        "length": 0,
        "sha1": b"\x00",
        "sha256": b"\x00",
        "sha1_git": b"\x00",
        "blake2s256": b"\x00",
    }
    with pytest.raises(ValueError, match="must be a timezone-aware datetime."):
        Content.from_dict(content_d)


@given(binary(max_size=4096))
def test_content_from_data(data):
    c = Content.from_data(data)
    assert c.data == data
    assert c.length == len(data)
    assert c.status == "visible"
    for key, value in MultiHash.from_data(data).digest().items():
        assert getattr(c, key) == value


@given(binary(max_size=4096))
def test_hidden_content_from_data(data):
    c = Content.from_data(data, status="hidden")
    assert c.data == data
    assert c.length == len(data)
    assert c.status == "hidden"
    for key, value in MultiHash.from_data(data).digest().items():
        assert getattr(c, key) == value


def test_content_naive_datetime():
    c = Content.from_data(b"foo")
    with pytest.raises(ValueError, match="must be a timezone-aware datetime"):
        Content(
            **c.to_dict(),
            ctime=datetime.datetime.now(),
        )


@given(strategies.present_contents())
def test_content_git_roundtrip(content):
    assert content.data is not None
    raw = swh.model.git_objects.content_git_object(content)
    sha1_git = hashlib.new("sha1", raw).digest()
    assert content.sha1_git == sha1_git


# SkippedContent


@given(binary(max_size=4096))
def test_skipped_content_from_data(data):
    c = SkippedContent.from_data(data, reason="reason")
    assert c.reason == "reason"
    assert c.length == len(data)
    assert c.status == "absent"
    for key, value in MultiHash.from_data(data).digest().items():
        assert getattr(c, key) == value


@given(strategies.skipped_contents_d())
def test_skipped_content_origin_is_str(skipped_content_d):
    assert SkippedContent.from_dict(skipped_content_d)

    skipped_content_d["origin"] = "http://path/to/origin"
    assert SkippedContent.from_dict(skipped_content_d)

    skipped_content_d["origin"] = Origin(url="http://path/to/origin")
    with pytest.raises(ValueError, match="origin"):
        SkippedContent.from_dict(skipped_content_d)


def test_skipped_content_naive_datetime():
    c = SkippedContent.from_data(b"foo", reason="reason")
    with pytest.raises(ValueError, match="must be a timezone-aware datetime"):
        SkippedContent(
            **c.to_dict(),
            ctime=datetime.datetime.now(),
        )


def test_skipped_content_swhid():
    skipped_content = SkippedContent.from_data(b"foo", reason="reason")
    assert skipped_content.swhid() == CoreSWHID.from_string(
        "swh:1:cnt:19102815663d23f8b75a47e7a01965dcdc96468c"
    )


# Directory


@given(strategies.directories(raw_manifest=none()))
def test_directory_check(directory):
    directory.check()

    directory2 = attr.evolve(directory, id=b"\x00" * 20)
    with pytest.raises(ValueError, match="does not match recomputed hash"):
        directory2.check()

    directory2 = attr.evolve(
        directory, raw_manifest=swh.model.git_objects.directory_git_object(directory)
    )
    with pytest.raises(
        ValueError, match="non-none raw_manifest attribute, but does not need it."
    ):
        directory2.check()


@given(strategies.directories(raw_manifest=none()))
def test_directory_raw_manifest(directory):
    assert "raw_manifest" not in directory.to_dict()

    raw_manifest = b"foo"
    id_ = hashlib.new("sha1", raw_manifest).digest()

    directory2 = attr.evolve(directory, raw_manifest=raw_manifest)
    assert directory2.to_dict()["raw_manifest"] == raw_manifest
    with pytest.raises(ValueError, match="does not match recomputed hash"):
        directory2.check()

    directory2 = attr.evolve(directory, raw_manifest=raw_manifest, id=id_)
    assert directory2.id is not None
    assert directory2.id == id_ != directory.id
    assert directory2.to_dict()["raw_manifest"] == raw_manifest
    directory2.check()


def test_directory_entry_name_validation():
    with pytest.raises(ValueError, match="valid directory entry name."):
        DirectoryEntry(name=b"foo/", type="dir", target=b"\x00" * 20, perms=0),


def test_directory_duplicate_entry_name():
    entries = (
        DirectoryEntry(name=b"foo", type="file", target=b"\x00" * 20, perms=0),
        DirectoryEntry(name=b"foo", type="dir", target=b"\x01" * 20, perms=1),
    )
    with pytest.raises(ValueError, match="duplicated entry name"):
        Directory(entries=entries)

    entries = (
        DirectoryEntry(name=b"foo", type="file", target=b"\x00" * 20, perms=0),
        DirectoryEntry(name=b"foo", type="file", target=b"\x00" * 20, perms=0),
    )
    with pytest.raises(ValueError, match="duplicated entry name"):
        Directory(entries=entries)


@given(strategies.directories())
def test_directory_from_possibly_duplicated_entries__no_duplicates(directory):
    """
    Directory.from_possibly_duplicated_entries should return the directory
    unchanged if it has no duplicated entry name.
    """
    assert (False, directory) == Directory.from_possibly_duplicated_entries(
        id=directory.id, entries=directory.entries, raw_manifest=directory.raw_manifest
    )
    assert (False, directory) == Directory.from_possibly_duplicated_entries(
        entries=directory.entries, raw_manifest=directory.raw_manifest
    )


@pytest.mark.parametrize("rev_first", [True, False])
def test_directory_from_possibly_duplicated_entries__rev_and_dir(rev_first):
    entries = (
        DirectoryEntry(name=b"foo", type="dir", target=b"\x01" * 20, perms=1),
        DirectoryEntry(name=b"foo", type="rev", target=b"\x00" * 20, perms=0),
    )
    if rev_first:
        entries = tuple(reversed(entries))
    (is_corrupt, dir_) = Directory.from_possibly_duplicated_entries(entries=entries)
    assert is_corrupt
    assert dir_.entries == (
        DirectoryEntry(name=b"foo", type="rev", target=b"\x00" * 20, perms=0),
        DirectoryEntry(
            name=b"foo_0101010101", type="dir", target=b"\x01" * 20, perms=1
        ),
    )

    # order is independent of 'rev_first' because it is always sorted in git order
    assert dir_.raw_manifest == (
        # fmt: off
        b"tree 52\x00"
        + b"0 foo\x00" + b"\x00" * 20
        + b"1 foo\x00" + b"\x01" * 20
        # fmt: on
    )


@pytest.mark.parametrize("file_first", [True, False])
def test_directory_from_possibly_duplicated_entries__file_and_dir(file_first):
    entries = (
        DirectoryEntry(name=b"foo", type="dir", target=b"\x01" * 20, perms=1),
        DirectoryEntry(name=b"foo", type="file", target=b"\x00" * 20, perms=0),
    )
    if file_first:
        entries = tuple(reversed(entries))
    (is_corrupt, dir_) = Directory.from_possibly_duplicated_entries(entries=entries)
    assert is_corrupt
    assert dir_.entries == (
        DirectoryEntry(name=b"foo", type="dir", target=b"\x01" * 20, perms=1),
        DirectoryEntry(
            name=b"foo_0000000000", type="file", target=b"\x00" * 20, perms=0
        ),
    )

    # order is independent of 'file_first' because it is always sorted in git order
    assert dir_.raw_manifest == (
        # fmt: off
        b"tree 52\x00"
        + b"0 foo\x00" + b"\x00" * 20
        + b"1 foo\x00" + b"\x01" * 20
        # fmt: on
    )


def test_directory_from_possibly_duplicated_entries__two_files1():
    entries = (
        DirectoryEntry(name=b"foo", type="file", target=b"\x01" * 20, perms=1),
        DirectoryEntry(name=b"foo", type="file", target=b"\x00" * 20, perms=0),
    )
    (is_corrupt, dir_) = Directory.from_possibly_duplicated_entries(entries=entries)
    assert is_corrupt

    assert dir_.entries == (
        DirectoryEntry(name=b"foo", type="file", target=b"\x01" * 20, perms=1),
        DirectoryEntry(
            name=b"foo_0000000000", type="file", target=b"\x00" * 20, perms=0
        ),
    )
    assert dir_.raw_manifest == (
        # fmt: off
        b"tree 52\x00"
        + b"1 foo\x00" + b"\x01" * 20
        + b"0 foo\x00" + b"\x00" * 20
        # fmt: on
    )


def test_directory_from_possibly_duplicated_entries__two_files2():
    """
    Same as above, but entries are in a different order (and order matters
    to break the tie)
    """
    entries = (
        DirectoryEntry(name=b"foo", type="file", target=b"\x00" * 20, perms=0),
        DirectoryEntry(name=b"foo", type="file", target=b"\x01" * 20, perms=1),
    )
    (is_corrupt, dir_) = Directory.from_possibly_duplicated_entries(entries=entries)
    assert is_corrupt

    assert dir_.entries == (
        DirectoryEntry(name=b"foo", type="file", target=b"\x00" * 20, perms=0),
        DirectoryEntry(
            name=b"foo_0101010101", type="file", target=b"\x01" * 20, perms=1
        ),
    )
    assert dir_.raw_manifest == (
        # fmt: off
        b"tree 52\x00"
        + b"0 foo\x00" + b"\x00" * 20
        + b"1 foo\x00" + b"\x01" * 20
        # fmt: on
    )


def test_directory_from_possibly_duplicated_entries__preserve_manifest():
    entries = (
        DirectoryEntry(name=b"foo", type="dir", target=b"\x01" * 20, perms=1),
        DirectoryEntry(name=b"foo", type="rev", target=b"\x00" * 20, perms=0),
    )
    (is_corrupt, dir_) = Directory.from_possibly_duplicated_entries(
        entries=entries, raw_manifest=b"blah"
    )
    assert is_corrupt
    assert dir_.entries == (
        DirectoryEntry(name=b"foo", type="rev", target=b"\x00" * 20, perms=0),
        DirectoryEntry(
            name=b"foo_0101010101", type="dir", target=b"\x01" * 20, perms=1
        ),
    )

    assert dir_.raw_manifest == b"blah"


@pytest.fixture
def directory_with_every_possible_type():
    return Directory.from_dict(
        {
            "entries": [
                {
                    "type": "file",
                    "perms": 33188,
                    "name": b"README",
                    "target": hash_to_bytes("37ec8ea2110c0b7a32fbb0e872f6e7debbf95e21"),
                },
                {
                    "type": "dir",
                    "perms": 16384,
                    "name": b"src",
                    "target": hash_to_bytes("61e6e867f5d7ba3b40540869bc050b0c4fed9e95"),
                },
                {
                    "type": "rev",
                    "perms": 57344,
                    "name": b"submodule",
                    "target": hash_to_bytes("3d531e169db92a16a9a8974f0ae6edf52e52659e"),
                },
            ],
        }
    )


def test_directory_entry_swhids(directory_with_every_possible_type):
    assert [entry.swhid() for entry in directory_with_every_possible_type.entries] == [
        CoreSWHID.from_string("swh:1:cnt:37ec8ea2110c0b7a32fbb0e872f6e7debbf95e21"),
        CoreSWHID.from_string("swh:1:dir:61e6e867f5d7ba3b40540869bc050b0c4fed9e95"),
        CoreSWHID.from_string("swh:1:rev:3d531e169db92a16a9a8974f0ae6edf52e52659e"),
    ]


# Release


@given(strategies.releases(raw_manifest=none()))
def test_release_check(release):
    release.check()

    release2 = attr.evolve(release, id=b"\x00" * 20)
    with pytest.raises(ValueError, match="does not match recomputed hash"):
        release2.check()

    release2 = attr.evolve(
        release, raw_manifest=swh.model.git_objects.release_git_object(release)
    )
    with pytest.raises(
        ValueError, match="non-none raw_manifest attribute, but does not need it."
    ):
        release2.check()


@given(strategies.releases(raw_manifest=none()))
def test_release_raw_manifest(release):
    raw_manifest = b"foo"
    id_ = hashlib.new("sha1", raw_manifest).digest()

    release2 = attr.evolve(release, raw_manifest=raw_manifest)
    assert release2.to_dict()["raw_manifest"] == raw_manifest
    with pytest.raises(ValueError, match="does not match recomputed hash"):
        release2.check()

    release2 = attr.evolve(release, raw_manifest=raw_manifest, id=id_)
    assert release2.id is not None
    assert release2.id == id_ != release.id
    assert release2.to_dict()["raw_manifest"] == raw_manifest
    release2.check()


def test_release_target_swhid():
    release = Release.from_dict(release_example)
    assert release.target_swhid() == CoreSWHID.from_string(
        "swh:1:rev:741b2252a5e14d6c60a913c77a6099abe73a854a"
    )


# Revision


@given(strategies.revisions(raw_manifest=none()))
def test_revision_check(revision):
    revision.check()

    revision2 = attr.evolve(revision, id=b"\x00" * 20)
    with pytest.raises(ValueError, match="does not match recomputed hash"):
        revision2.check()

    revision2 = attr.evolve(
        revision, raw_manifest=swh.model.git_objects.revision_git_object(revision)
    )
    with pytest.raises(
        ValueError, match="non-none raw_manifest attribute, but does not need it."
    ):
        revision2.check()


@given(strategies.revisions(raw_manifest=none()))
def test_revision_raw_manifest(revision):

    raw_manifest = b"foo"
    id_ = hashlib.new("sha1", raw_manifest).digest()

    revision2 = attr.evolve(revision, raw_manifest=raw_manifest)
    assert revision2.to_dict()["raw_manifest"] == raw_manifest
    with pytest.raises(ValueError, match="does not match recomputed hash"):
        revision2.check()

    revision2 = attr.evolve(revision, raw_manifest=raw_manifest, id=id_)
    assert revision2.id is not None
    assert revision2.id == id_ != revision.id
    assert revision2.to_dict()["raw_manifest"] == raw_manifest
    revision2.check()


def test_revision_extra_headers_no_headers():
    rev_dict = revision_example.copy()
    rev_dict.pop("id")
    rev = Revision.from_dict(rev_dict)
    rev_dict = attr.asdict(rev, recurse=False)

    rev_model = Revision(**rev_dict)
    assert rev_model.metadata is None
    assert rev_model.extra_headers == ()

    rev_dict["metadata"] = {
        "something": "somewhere",
        "some other thing": "stranger",
    }
    rev_model = Revision(**rev_dict)
    assert rev_model.metadata == rev_dict["metadata"]
    assert rev_model.extra_headers == ()


def test_revision_extra_headers_with_headers():
    rev_dict = revision_example.copy()
    rev_dict.pop("id")
    rev = Revision.from_dict(rev_dict)
    rev_dict = attr.asdict(rev, recurse=False)
    rev_dict["metadata"] = {
        "something": "somewhere",
        "some other thing": "stranger",
    }
    extra_headers = (
        (b"header1", b"value1"),
        (b"header2", b"42"),
        (b"header3", b"should I?\x00"),
        (b"header1", b"again"),
    )

    rev_dict["extra_headers"] = extra_headers
    rev_model = Revision(**rev_dict)
    assert "extra_headers" not in rev_model.metadata
    assert rev_model.extra_headers == extra_headers


def test_revision_extra_headers_in_metadata():
    rev_dict = revision_example.copy()
    rev_dict.pop("id")
    rev = Revision.from_dict(rev_dict)
    rev_dict = attr.asdict(rev, recurse=False)
    rev_dict["metadata"] = {
        "something": "somewhere",
        "some other thing": "stranger",
    }

    extra_headers = (
        (b"header1", b"value1"),
        (b"header2", b"42"),
        (b"header3", b"should I?\x00"),
        (b"header1", b"again"),
    )

    # check the bw-compat init hook does the job
    # ie. extra_headers are given in the metadata field
    rev_dict["metadata"]["extra_headers"] = extra_headers
    rev_model = Revision(**rev_dict)
    assert "extra_headers" not in rev_model.metadata
    assert rev_model.extra_headers == extra_headers


def test_revision_extra_headers_as_lists():
    rev_dict = revision_example.copy()
    rev_dict.pop("id")
    rev = Revision.from_dict(rev_dict)
    rev_dict = attr.asdict(rev, recurse=False)
    rev_dict["metadata"] = {}

    extra_headers = (
        (b"header1", b"value1"),
        (b"header2", b"42"),
        (b"header3", b"should I?\x00"),
        (b"header1", b"again"),
    )

    # check Revision.extra_headers tuplify does the job
    rev_dict["extra_headers"] = [list(x) for x in extra_headers]
    rev_model = Revision(**rev_dict)
    assert "extra_headers" not in rev_model.metadata
    assert rev_model.extra_headers == extra_headers


def test_revision_extra_headers_type_error():
    rev_dict = revision_example.copy()
    rev_dict.pop("id")
    rev = Revision.from_dict(rev_dict)
    orig_rev_dict = attr.asdict(rev, recurse=False)
    orig_rev_dict["metadata"] = {
        "something": "somewhere",
        "some other thing": "stranger",
    }
    extra_headers = (
        ("header1", b"value1"),
        (b"header2", 42),
        ("header1", "again"),
    )
    # check headers one at a time
    #   if given as extra_header
    for extra_header in extra_headers:
        rev_dict = copy.deepcopy(orig_rev_dict)
        rev_dict["extra_headers"] = (extra_header,)
        with pytest.raises(AttributeTypeError):
            Revision(**rev_dict)
    #   if given as metadata
    for extra_header in extra_headers:
        rev_dict = copy.deepcopy(orig_rev_dict)
        rev_dict["metadata"]["extra_headers"] = (extra_header,)
        with pytest.raises(AttributeTypeError):
            Revision(**rev_dict)


def test_revision_extra_headers_from_dict():
    rev_dict = revision_example.copy()
    rev_dict.pop("id")
    rev_model = Revision.from_dict(rev_dict)
    assert rev_model.metadata is None
    assert rev_model.extra_headers == ()

    rev_dict["metadata"] = {
        "something": "somewhere",
        "some other thing": "stranger",
    }
    rev_model = Revision.from_dict(rev_dict)
    assert rev_model.metadata == rev_dict["metadata"]
    assert rev_model.extra_headers == ()

    extra_headers = (
        (b"header1", b"value1"),
        (b"header2", b"42"),
        (b"header3", b"should I?\nmaybe\x00\xff"),
        (b"header1", b"again"),
    )
    rev_dict["extra_headers"] = extra_headers
    rev_model = Revision.from_dict(rev_dict)
    assert "extra_headers" not in rev_model.metadata
    assert rev_model.extra_headers == extra_headers


def test_revision_extra_headers_in_metadata_from_dict():
    rev_dict = revision_example.copy()
    rev_dict.pop("id")

    rev_dict["metadata"] = {
        "something": "somewhere",
        "some other thing": "stranger",
    }
    extra_headers = (
        (b"header1", b"value1"),
        (b"header2", b"42"),
        (b"header3", b"should I?\nmaybe\x00\xff"),
        (b"header1", b"again"),
    )
    # check the bw-compat init hook does the job
    rev_dict["metadata"]["extra_headers"] = extra_headers
    rev_model = Revision.from_dict(rev_dict)
    assert "extra_headers" not in rev_model.metadata
    assert rev_model.extra_headers == extra_headers


def test_revision_extra_headers_as_lists_from_dict():
    rev_dict = revision_example.copy()
    rev_dict.pop("id")
    rev_model = Revision.from_dict(rev_dict)
    rev_dict["metadata"] = {
        "something": "somewhere",
        "some other thing": "stranger",
    }
    extra_headers = (
        (b"header1", b"value1"),
        (b"header2", b"42"),
        (b"header3", b"should I?\nmaybe\x00\xff"),
        (b"header1", b"again"),
    )
    # check Revision.extra_headers converter does the job
    rev_dict["extra_headers"] = [list(x) for x in extra_headers]
    rev_model = Revision.from_dict(rev_dict)
    assert "extra_headers" not in rev_model.metadata
    assert rev_model.extra_headers == extra_headers


def test_revision_no_author_or_committer_from_dict():
    rev_dict = revision_example.copy()
    rev_dict["author"] = rev_dict["date"] = None
    rev_dict["committer"] = rev_dict["committer_date"] = None
    rev_model = Revision.from_dict(rev_dict)
    assert rev_model.to_dict() == {
        **rev_dict,
        "parents": tuple(rev_dict["parents"]),
        "extra_headers": (),
        "metadata": None,
    }


def test_revision_none_author_or_committer():
    rev_dict = revision_example.copy()
    rev_dict["author"] = None
    with pytest.raises(ValueError, match=".*date must be None if author is None.*"):
        Revision.from_dict(rev_dict)

    rev_dict = revision_example.copy()
    rev_dict["committer"] = None
    with pytest.raises(
        ValueError, match=".*committer_date must be None if committer is None.*"
    ):
        Revision.from_dict(rev_dict)


def test_revision_directory_swhid():
    revision = Revision.from_dict(revision_example)
    assert revision.directory_swhid() == CoreSWHID.from_string(
        "swh:1:dir:85a74718d377195e1efd0843ba4f3260bad4fe07"
    )


def test_revision_parent_swhids():
    revision_d = copy.deepcopy(revision_example)
    revision_d["parents"].append(
        hash_to_bytes("b2a7e1260492e344fab3cbf91bc13c91e05426fd")
    )
    revision = Revision.from_dict(revision_d)
    assert revision.parent_swhids() == [
        CoreSWHID.from_string("swh:1:rev:01e2d0627a9a6edb24c37db45db5ecb31e9de808"),
        CoreSWHID.from_string("swh:1:rev:b2a7e1260492e344fab3cbf91bc13c91e05426fd"),
    ]


@pytest.fixture
def snapshot_with_all_types():
    return Snapshot.from_dict(snapshot_example)


def test_snapshot_branch_swhids(snapshot_with_all_types):
    assert {
        name: branch and branch.swhid()
        for (name, branch) in snapshot_with_all_types.branches.items()
    } == {
        b"directory": CoreSWHID.from_string(
            "swh:1:dir:1bd0e65f7d2ff14ae994de17a1e7fe65111dcad8"
        ),
        b"content": CoreSWHID.from_string(
            "swh:1:cnt:fe95a46679d128ff167b7c55df5d02356c5a1ae1"
        ),
        b"alias": None,
        b"revision": CoreSWHID.from_string(
            "swh:1:rev:aafb16d69fd30ff58afdd69036a26047f3aebdc6"
        ),
        b"release": CoreSWHID.from_string(
            "swh:1:rel:7045404f3d1c54e6473c71bbb716529fbad4be24"
        ),
        b"snapshot": CoreSWHID.from_string(
            "swh:1:snp:1a8893e6a86f444e8be8e7bda6cb34fb1735a00e"
        ),
        b"dangling": None,
    }


@given(strategies.objects(split_content=True))
def test_object_type(objtype_and_obj):
    obj_type, obj = objtype_and_obj
    assert obj_type == obj.object_type


def test_object_type_is_final():
    object_types = set()

    def check_final(cls):
        if hasattr(cls, "object_type"):
            assert cls.object_type not in object_types
            object_types.add(cls.object_type)
        if cls.__subclasses__():
            assert not hasattr(cls, "object_type")
        for subcls in cls.__subclasses__():
            check_final(subcls)

    check_final(BaseModel)


_metadata_authority = MetadataAuthority(
    type=MetadataAuthorityType.FORGE,
    url="https://forge.softwareheritage.org",
)
_metadata_fetcher = MetadataFetcher(
    name="test-fetcher",
    version="0.0.1",
)
_content_swhid = ExtendedSWHID.from_string(
    "swh:1:cnt:94a9ed024d3859793618152ea559a168bbcbb5e2"
)
_origin_url = "https://forge.softwareheritage.org/source/swh-model.git"
_origin_swhid = ExtendedSWHID.from_string(
    "swh:1:ori:433b4f5612f0720ed51fa7aeaf43a3625870057b"
)
_dummy_qualifiers = {"origin": "https://example.com", "lines": "42"}
_common_metadata_fields = dict(
    discovery_date=datetime.datetime(
        2021, 1, 29, 13, 57, 9, tzinfo=datetime.timezone.utc
    ),
    authority=_metadata_authority,
    fetcher=_metadata_fetcher,
    format="json",
    metadata=b'{"origin": "https://example.com", "lines": "42"}',
)


def test_metadata_valid():
    """Checks valid RawExtrinsicMetadata objects don't raise an error."""

    # Simplest case
    RawExtrinsicMetadata(target=_origin_swhid, **_common_metadata_fields)

    # Object with an SWHID
    RawExtrinsicMetadata(
        target=_content_swhid,
        **_common_metadata_fields,
    )


def test_metadata_from_old_dict():
    common_fields = {
        "authority": {"type": "forge", "url": "https://forge.softwareheritage.org"},
        "fetcher": {
            "name": "test-fetcher",
            "version": "0.0.1",
        },
        "discovery_date": _common_metadata_fields["discovery_date"],
        "format": "json",
        "metadata": b'{"origin": "https://example.com", "lines": "42"}',
    }

    m = RawExtrinsicMetadata(
        target=_origin_swhid,
        **_common_metadata_fields,
    )
    assert (
        RawExtrinsicMetadata.from_dict(
            {"id": m.id, "target": _origin_url, "type": "origin", **common_fields}
        )
        == m
    )

    m = RawExtrinsicMetadata(
        target=_content_swhid,
        **_common_metadata_fields,
    )
    assert (
        RawExtrinsicMetadata.from_dict(
            {"target": str(_content_swhid), "type": "content", **common_fields}
        )
        == m
    )


def test_metadata_to_dict():
    """Checks valid RawExtrinsicMetadata objects don't raise an error."""

    common_fields = {
        "authority": {"type": "forge", "url": "https://forge.softwareheritage.org"},
        "fetcher": {
            "name": "test-fetcher",
            "version": "0.0.1",
        },
        "discovery_date": _common_metadata_fields["discovery_date"],
        "format": "json",
        "metadata": b'{"origin": "https://example.com", "lines": "42"}',
    }

    m = RawExtrinsicMetadata(
        target=_origin_swhid,
        **_common_metadata_fields,
    )
    assert m.to_dict() == {
        "target": str(_origin_swhid),
        "id": b"\xa3)q\x0f\xf7p\xc7\xb0\\O\xe8\x84\x83Z\xb0]\x81\xe9\x95\x13",
        **common_fields,
    }
    assert RawExtrinsicMetadata.from_dict(m.to_dict()) == m

    m = RawExtrinsicMetadata(
        target=_content_swhid,
        **_common_metadata_fields,
    )
    assert m.to_dict() == {
        "target": "swh:1:cnt:94a9ed024d3859793618152ea559a168bbcbb5e2",
        "id": b"\xbc\xa3U\xddf\x19U\xc5\xd2\xd7\xdfK\xd7c\x1f\xa8\xfeh\x992",
        **common_fields,
    }
    assert RawExtrinsicMetadata.from_dict(m.to_dict()) == m

    hash_hex = "6162" * 10
    hash_bin = b"ab" * 10
    m = RawExtrinsicMetadata(
        target=_content_swhid,
        **_common_metadata_fields,
        origin="https://example.org/",
        snapshot=CoreSWHID(object_type=ObjectType.SNAPSHOT, object_id=hash_bin),
        release=CoreSWHID(object_type=ObjectType.RELEASE, object_id=hash_bin),
        revision=CoreSWHID(object_type=ObjectType.REVISION, object_id=hash_bin),
        path=b"/foo/bar",
        directory=CoreSWHID(object_type=ObjectType.DIRECTORY, object_id=hash_bin),
    )
    assert m.to_dict() == {
        "target": "swh:1:cnt:94a9ed024d3859793618152ea559a168bbcbb5e2",
        "id": b"\x14l\xb0\x1f\xb9\xc0{)\xc7\x0f\xbd\xc0*,YZ\xf5C\xab\xfc",
        **common_fields,
        "origin": "https://example.org/",
        "snapshot": f"swh:1:snp:{hash_hex}",
        "release": f"swh:1:rel:{hash_hex}",
        "revision": f"swh:1:rev:{hash_hex}",
        "path": b"/foo/bar",
        "directory": f"swh:1:dir:{hash_hex}",
    }
    assert RawExtrinsicMetadata.from_dict(m.to_dict()) == m


def test_metadata_invalid_target():
    """Checks various invalid values for the 'target' field."""
    # SWHID passed as string instead of SWHID
    with pytest.raises(ValueError, match="target must be.*ExtendedSWHID"):
        RawExtrinsicMetadata(
            target="swh:1:cnt:94a9ed024d3859793618152ea559a168bbcbb5e2",
            **_common_metadata_fields,
        )


def test_metadata_naive_datetime():
    with pytest.raises(ValueError, match="must be a timezone-aware datetime"):
        RawExtrinsicMetadata(
            target=_origin_swhid,
            **{**_common_metadata_fields, "discovery_date": datetime.datetime.now()},
        )


def test_metadata_validate_context_origin():
    """Checks validation of RawExtrinsicMetadata.origin."""

    # Origins can't have an 'origin' context
    with pytest.raises(
        ValueError, match="Unexpected 'origin' context for origin object"
    ):
        RawExtrinsicMetadata(
            target=_origin_swhid,
            origin=_origin_url,
            **_common_metadata_fields,
        )

    # but all other types can
    RawExtrinsicMetadata(
        target=_content_swhid,
        origin=_origin_url,
        **_common_metadata_fields,
    )

    # SWHIDs aren't valid origin URLs
    with pytest.raises(ValueError, match="SWHID used as context origin URL"):
        RawExtrinsicMetadata(
            target=_content_swhid,
            origin="swh:1:cnt:94a9ed024d3859793618152ea559a168bbcbb5e2",
            **_common_metadata_fields,
        )


def test_metadata_validate_context_visit():
    """Checks validation of RawExtrinsicMetadata.visit."""

    # Origins can't have a 'visit' context
    with pytest.raises(
        ValueError, match="Unexpected 'visit' context for origin object"
    ):
        RawExtrinsicMetadata(
            target=_origin_swhid,
            visit=42,
            **_common_metadata_fields,
        )

    # but all other types can
    RawExtrinsicMetadata(
        target=_content_swhid,
        origin=_origin_url,
        visit=42,
        **_common_metadata_fields,
    )

    # Missing 'origin'
    with pytest.raises(ValueError, match="'origin' context must be set if 'visit' is"):
        RawExtrinsicMetadata(
            target=_content_swhid,
            visit=42,
            **_common_metadata_fields,
        )

    # visit id must be positive
    with pytest.raises(ValueError, match="Nonpositive visit id"):
        RawExtrinsicMetadata(
            target=_content_swhid,
            origin=_origin_url,
            visit=-42,
            **_common_metadata_fields,
        )


def test_metadata_validate_context_snapshot():
    """Checks validation of RawExtrinsicMetadata.snapshot."""

    # Origins can't have a 'snapshot' context
    with pytest.raises(
        ValueError, match="Unexpected 'snapshot' context for origin object"
    ):
        RawExtrinsicMetadata(
            target=_origin_swhid,
            snapshot=CoreSWHID(
                object_type=ObjectType.SNAPSHOT,
                object_id=EXAMPLE_HASH,
            ),
            **_common_metadata_fields,
        )

    # but content can
    RawExtrinsicMetadata(
        target=_content_swhid,
        snapshot=CoreSWHID(object_type=ObjectType.SNAPSHOT, object_id=EXAMPLE_HASH),
        **_common_metadata_fields,
    )

    # SWHID type doesn't match the expected type of this context key
    with pytest.raises(
        ValueError, match="Expected SWHID type 'snapshot', got 'content'"
    ):
        RawExtrinsicMetadata(
            target=_content_swhid,
            snapshot=CoreSWHID(
                object_type=ObjectType.CONTENT,
                object_id=EXAMPLE_HASH,
            ),
            **_common_metadata_fields,
        )


def test_metadata_validate_context_release():
    """Checks validation of RawExtrinsicMetadata.release."""

    # Origins can't have a 'release' context
    with pytest.raises(
        ValueError, match="Unexpected 'release' context for origin object"
    ):
        RawExtrinsicMetadata(
            target=_origin_swhid,
            release=CoreSWHID(
                object_type=ObjectType.RELEASE,
                object_id=EXAMPLE_HASH,
            ),
            **_common_metadata_fields,
        )

    # but content can
    RawExtrinsicMetadata(
        target=_content_swhid,
        release=CoreSWHID(object_type=ObjectType.RELEASE, object_id=EXAMPLE_HASH),
        **_common_metadata_fields,
    )

    # SWHID type doesn't match the expected type of this context key
    with pytest.raises(
        ValueError, match="Expected SWHID type 'release', got 'content'"
    ):
        RawExtrinsicMetadata(
            target=_content_swhid,
            release=CoreSWHID(
                object_type=ObjectType.CONTENT,
                object_id=EXAMPLE_HASH,
            ),
            **_common_metadata_fields,
        )


def test_metadata_validate_context_revision():
    """Checks validation of RawExtrinsicMetadata.revision."""

    # Origins can't have a 'revision' context
    with pytest.raises(
        ValueError, match="Unexpected 'revision' context for origin object"
    ):
        RawExtrinsicMetadata(
            target=_origin_swhid,
            revision=CoreSWHID(
                object_type=ObjectType.REVISION,
                object_id=EXAMPLE_HASH,
            ),
            **_common_metadata_fields,
        )

    # but content can
    RawExtrinsicMetadata(
        target=_content_swhid,
        revision=CoreSWHID(object_type=ObjectType.REVISION, object_id=EXAMPLE_HASH),
        **_common_metadata_fields,
    )

    # SWHID type doesn't match the expected type of this context key
    with pytest.raises(
        ValueError, match="Expected SWHID type 'revision', got 'content'"
    ):
        RawExtrinsicMetadata(
            target=_content_swhid,
            revision=CoreSWHID(
                object_type=ObjectType.CONTENT,
                object_id=EXAMPLE_HASH,
            ),
            **_common_metadata_fields,
        )


def test_metadata_validate_context_path():
    """Checks validation of RawExtrinsicMetadata.path."""

    # Origins can't have a 'path' context
    with pytest.raises(ValueError, match="Unexpected 'path' context for origin object"):
        RawExtrinsicMetadata(
            target=_origin_swhid,
            path=b"/foo/bar",
            **_common_metadata_fields,
        )

    # but content can
    RawExtrinsicMetadata(
        target=_content_swhid,
        path=b"/foo/bar",
        **_common_metadata_fields,
    )


def test_metadata_validate_context_directory():
    """Checks validation of RawExtrinsicMetadata.directory."""

    # Origins can't have a 'directory' context
    with pytest.raises(
        ValueError, match="Unexpected 'directory' context for origin object"
    ):
        RawExtrinsicMetadata(
            target=_origin_swhid,
            directory=CoreSWHID(
                object_type=ObjectType.DIRECTORY,
                object_id=EXAMPLE_HASH,
            ),
            **_common_metadata_fields,
        )

    # but content can
    RawExtrinsicMetadata(
        target=_content_swhid,
        directory=CoreSWHID(
            object_type=ObjectType.DIRECTORY,
            object_id=EXAMPLE_HASH,
        ),
        **_common_metadata_fields,
    )

    # SWHID type doesn't match the expected type of this context key
    with pytest.raises(
        ValueError, match="Expected SWHID type 'directory', got 'content'"
    ):
        RawExtrinsicMetadata(
            target=_content_swhid,
            directory=CoreSWHID(
                object_type=ObjectType.CONTENT,
                object_id=EXAMPLE_HASH,
            ),
            **_common_metadata_fields,
        )


def test_metadata_normalize_discovery_date():
    fields_copy = {**_common_metadata_fields}
    truncated_date = fields_copy.pop("discovery_date")
    assert truncated_date.microsecond == 0

    # Check for TypeError on disabled object type: we removed attrs_strict's
    # type_validator
    with pytest.raises(TypeError):
        RawExtrinsicMetadata(
            target=_content_swhid, discovery_date="not a datetime", **fields_copy
        )

    # Check for truncation to integral second
    date_with_us = truncated_date.replace(microsecond=42)
    md = RawExtrinsicMetadata(
        target=_content_swhid,
        discovery_date=date_with_us,
        **fields_copy,
    )

    assert md.discovery_date == truncated_date
    assert md.discovery_date.tzinfo == datetime.timezone.utc

    # Check that the timezone gets normalized. Timezones can be offset by a
    # non-integral number of seconds, so we need to handle that.
    timezone = datetime.timezone(offset=datetime.timedelta(hours=2))
    date_with_tz = truncated_date.astimezone(timezone)

    assert date_with_tz.tzinfo != datetime.timezone.utc

    md = RawExtrinsicMetadata(
        target=_content_swhid,
        discovery_date=date_with_tz,
        **fields_copy,
    )

    assert md.discovery_date == truncated_date
    assert md.discovery_date.tzinfo == datetime.timezone.utc


def test_revision_repr():
    from swh.model.model import RevisionType  # noqa

    revision = Revision.from_dict(revision_example)
    rev_repr = repr(revision)

    assert rev_repr == (
        "Revision(message=b'Linux 4.2-rc2\\n', "
        "author=Person(fullname=b'Linus Torvalds <torvalds@linux-foundation.org>', "
        "name=b'Linus Torvalds', email=b'torvalds@linux-foundation.org'), "
        "committer=Person(fullname=b'Linus Torvalds <torvalds@linux-foundation.org>', "
        "name=b'Linus Torvalds', email=b'torvalds@linux-foundation.org'), "
        "date=TimestampWithTimezone(timestamp=Timestamp(seconds=1436739030, microseconds=0), "
        "offset_bytes=b'-0700'), "
        "committer_date=TimestampWithTimezone(timestamp=Timestamp(seconds=1436739030, "
        "microseconds=0), offset_bytes=b'-0700'), "
        "type=RevisionType.GIT, "
        "directory=hash_to_bytes('85a74718d377195e1efd0843ba4f3260bad4fe07'), "
        "synthetic=False, metadata=None, "
        "parents=(hash_to_bytes('01e2d0627a9a6edb24c37db45db5ecb31e9de808'),), "
        "id=hash_to_bytes('bc0195aad0daa2ad5b0d76cce22b167bc3435590'), "
        "extra_headers=(), raw_manifest=None)"
    )
    assert eval(rev_repr) == revision



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Primitives for finding unknown content efficiently."""

from __future__ import annotations

from collections import namedtuple
import itertools
import logging
import random
from typing import Any, Iterable, List, Mapping, NamedTuple, Set, Union

from typing_extensions import Protocol, runtime_checkable

from .from_disk import model
from .model import Sha1Git

logger = logging.getLogger(__name__)

# Maximum amount when sampling from the undecided set of directory entries
SAMPLE_SIZE = 1000

# Sets of sha1 of contents, skipped contents and directories respectively
Sample: NamedTuple = namedtuple(
    "Sample", ["contents", "skipped_contents", "directories"]
)


@runtime_checkable
class ArchiveDiscoveryInterface(Protocol):
    """Interface used in discovery code to abstract over ways of connecting to
    the SWH archive (direct storage, web API, etc.) for all methods needed by
    discovery algorithms."""

    contents: List[model.Content]
    skipped_contents: List[model.SkippedContent]
    directories: List[model.Directory]

    def __init__(
        self,
        contents: List[model.Content],
        skipped_contents: List[model.SkippedContent],
        directories: List[model.Directory],
    ) -> None:
        self.contents = contents
        self.skipped_contents = skipped_contents
        self.directories = directories

    async def content_missing(self, contents: List[Sha1Git]) -> Iterable[Sha1Git]:
        """List content missing from the archive by sha1"""

    async def skipped_content_missing(
        self, skipped_contents: List[Sha1Git]
    ) -> Iterable[Sha1Git]:
        """List skipped content missing from the archive by sha1"""

    async def directory_missing(self, directories: List[Sha1Git]) -> Iterable[Sha1Git]:
        """List directories missing from the archive by sha1"""


class BaseDiscoveryGraph:
    """Creates the base structures and methods needed for discovery algorithms.
    Subclasses should override ``get_sample`` to affect how the discovery is made."""

    def __init__(self, contents, skipped_contents, directories):
        self._all_contents: Mapping[
            Sha1Git, Union[model.Content, model.SkippedContent]
        ] = {}
        self._undecided_directories: Set[Sha1Git] = set()
        self._children: Mapping[Sha1Git, model.DirectoryEntry] = {}
        self._parents: Mapping[model.DirectoryEntry, Sha1Git] = {}
        self.undecided: Set[Sha1Git] = set()

        for content in itertools.chain(contents, skipped_contents):
            self.undecided.add(content.sha1_git)
            self._all_contents[content.sha1_git] = content

        for directory in directories:
            self.undecided.add(directory.id)
            self._undecided_directories.add(directory.id)
            self._children[directory.id] = {c.target for c in directory.entries}
            for child in directory.entries:
                self._parents.setdefault(child.target, set()).add(directory.id)

        self.undecided |= self._undecided_directories
        self.known: Set[Sha1Git] = set()
        self.unknown: Set[Sha1Git] = set()

    def mark_known(self, entries: Iterable[Sha1Git]):
        """Mark ``entries`` and those they imply as known in the SWH archive"""
        self._mark_entries(entries, self._children, self.known)

    def mark_unknown(self, entries: Iterable[Sha1Git]):
        """Mark ``entries`` and those they imply as unknown in the SWH archive"""
        self._mark_entries(entries, self._parents, self.unknown)

    def _mark_entries(
        self,
        entries: Iterable[Sha1Git],
        transitive_mapping: Mapping[Any, Any],
        target_set: Set[Any],
    ):
        """Use Merkle graph properties to mark a directory entry as known or unknown.

        If an entry is known, then all of its descendants are known. If it's
        unknown, then all of its ancestors are unknown.

        - ``entries``: directory entries to mark along with their ancestors/descendants
          where applicable.
        - ``transitive_mapping``: mapping from an entry to the next entries to mark
          in the hierarchy, if any.
        - ``target_set``: set where marked entries will be added.

        """
        to_process = set(entries)
        while to_process:
            current = to_process.pop()
            target_set.add(current)
            self.undecided.discard(current)
            self._undecided_directories.discard(current)
            next_entries = transitive_mapping.get(current, set()) & self.undecided
            to_process.update(next_entries)

    async def get_sample(
        self,
    ) -> Sample:
        """Return a three-tuple of samples from the undecided sets of contents,
        skipped contents and directories respectively.
        These samples will be queried against the storage which will tell us
        which are known."""
        raise NotImplementedError()

    async def do_query(
        self, archive: ArchiveDiscoveryInterface, sample: Sample
    ) -> None:
        """Given a three-tuple of samples, ask the archive which are known or
        unknown and mark them as such."""

        methods = (
            archive.content_missing,
            archive.skipped_content_missing,
            archive.directory_missing,
        )

        for sample_per_type, method in zip(sample, methods):
            if not sample_per_type:
                continue
            known = set(sample_per_type)
            unknown = set(await method(list(sample_per_type)))
            known -= unknown

            self.mark_known(known)
            self.mark_unknown(unknown)


class RandomDirSamplingDiscoveryGraph(BaseDiscoveryGraph):
    """Use a random sampling using only directories.

    This allows us to find a statistically good spread of entries in the graph
    with a smaller population than using all types of entries. When there are
    no more directories, only contents or skipped contents are undecided if any
    are left: we send them directly to the storage since they should be few and
    their structure flat."""

    async def get_sample(self) -> Sample:
        if self._undecided_directories:
            if len(self._undecided_directories) <= SAMPLE_SIZE:
                return Sample(
                    contents=set(),
                    skipped_contents=set(),
                    directories=set(self._undecided_directories),
                )
            sample = random.sample(tuple(self._undecided_directories), SAMPLE_SIZE)
            directories = {o for o in sample}
            return Sample(
                contents=set(), skipped_contents=set(), directories=directories
            )

        contents = set()
        skipped_contents = set()

        for sha1 in self.undecided:
            obj = self._all_contents[sha1]
            obj_type = obj.object_type
            if obj_type == model.Content.object_type:
                contents.add(sha1)
            elif obj_type == model.SkippedContent.object_type:
                skipped_contents.add(sha1)
            else:
                raise TypeError(f"Unexpected object type {obj_type}")

        return Sample(
            contents=contents, skipped_contents=skipped_contents, directories=set()
        )


async def filter_known_objects(archive: ArchiveDiscoveryInterface):
    """Filter ``archive``'s ``contents``, ``skipped_contents`` and ``directories``
    to only return those that are unknown to the SWH archive using a discovery
    algorithm."""
    contents = archive.contents
    skipped_contents = archive.skipped_contents
    directories = archive.directories

    contents_count = len(contents)
    skipped_contents_count = len(skipped_contents)
    directories_count = len(directories)

    graph = RandomDirSamplingDiscoveryGraph(contents, skipped_contents, directories)

    while graph.undecided:
        sample = await graph.get_sample()
        await graph.do_query(archive, sample)

    contents = [c for c in contents if c.sha1_git in graph.unknown]
    skipped_contents = [c for c in skipped_contents if c.sha1_git in graph.unknown]
    directories = [c for c in directories if c.id in graph.unknown]

    logger.debug(
        "Filtered out %d contents, %d skipped contents and %d directories",
        contents_count - len(contents),
        skipped_contents_count - len(skipped_contents),
        directories_count - len(directories),
    )

    return (contents, skipped_contents, directories)



# Copyright (C) 2015-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import itertools

import attr
import pytest

from swh.model.exceptions import ValidationError
from swh.model.hashutil import hash_to_bytes as _x
from swh.model.swhids import (
    SWHID_QUALIFIERS,
    CoreSWHID,
    ExtendedObjectType,
    ExtendedSWHID,
    ObjectType,
    QualifiedSWHID,
)

dummy_qualifiers = {"origin": "https://example.com", "lines": "42"}


# SWHIDs that are outright invalid, no matter the context
INVALID_SWHIDS = [
    "swh:1:cnt",
    "swh:1:",
    "swh:",
    "swh:1:cnt:",
    "foo:1:cnt:abc8bc9d7a6bcf6db04f476d29314f157507d505",
    "swh:2:dir:def8bc9d7a6bcf6db04f476d29314f157507d505",
    "swh:1:foo:fed8bc9d7a6bcf6db04f476d29314f157507d505",
    "swh:1:dir:0b6959356d30f1a4e9b7f6bca59b9a336464c03d;invalid;malformed",
    "swh:1:snp:gh6959356d30f1a4e9b7f6bca59b9a336464c03d",
    "swh:1:snp:foo",
    # wrong qualifier: ori should be origin
    "swh:1:dir:0b6959356d30f1a4e9b7f6bca59b9a336464c03d;ori=something;anchor=1;visit=1;path=/",  # noqa
    # wrong qualifier: anc should be anchor
    "swh:1:dir:0b6959356d30f1a4e9b7f6bca59b9a336464c03d;origin=something;anc=1;visit=1;path=/",  # noqa
    # wrong qualifier: vis should be visit
    "swh:1:dir:0b6959356d30f1a4e9b7f6bca59b9a336464c03d;origin=something;anchor=1;vis=1;path=/",  # noqa
    # wrong qualifier: pa should be path
    "swh:1:dir:0b6959356d30f1a4e9b7f6bca59b9a336464c03d;origin=something;anchor=1;visit=1;pa=/",  # noqa
    # wrong qualifier: line should be lines
    "swh:1:dir:0b6959356d30f1a4e9b7f6bca59b9a336464c03d;line=10;origin=something;anchor=1;visit=1;path=/",  # noqa
    # wrong qualifier value: it contains space before of after
    "swh:1:dir:0b6959356d30f1a4e9b7f6bca59b9a336464c03d;origin=  https://some-url",  # noqa
    "swh:1:dir:0b6959356d30f1a4e9b7f6bca59b9a336464c03d;origin=something;anchor=some-anchor    ",  # noqa
    "swh:1:dir:0b6959356d30f1a4e9b7f6bca59b9a336464c03d;origin=something;anchor=some-anchor    ;visit=1",  # noqa
    # invalid swhid: whitespaces
    "swh :1:dir:0b6959356d30f1a4e9b7f6bca59b9a336464c03d;ori=something;anchor=1;visit=1;path=/",  # noqa
    "swh: 1:dir:0b6959356d30f1a4e9b7f6bca59b9a336464c03d;ori=something;anchor=1;visit=1;path=/",  # noqa
    "swh: 1: dir:0b6959356d30f1a4e9b7f6bca59b9a336464c03d;ori=something;anchor=1;visit=1;path=/",  # noqa
    "swh:1: dir: 0b6959356d30f1a4e9b7f6bca59b9a336464c03d",
    "swh:1: dir: 0b6959356d30f1a4e9b7f6bca59b9a336464c03d; origin=blah",
    "swh:1: dir: 0b6959356d30f1a4e9b7f6bca59b9a336464c03d;lines=12",
    # other whitespaces
    "swh\t:1:dir:0b6959356d30f1a4e9b7f6bca59b9a336464c03d;lines=12",
    "swh:1\n:dir:0b6959356d30f1a4e9b7f6bca59b9a336464c03d;lines=12",
    "swh:1:\rdir:0b6959356d30f1a4e9b7f6bca59b9a336464c03d;lines=12",
    "swh:1:dir:0b6959356d30f1a4e9b7f6bca59b9a336464c03d\f;lines=12",
    "swh:1:dir:0b6959356d30f1a4e9b7f6bca59b9a336464c03d;lines=12\v",
]

SWHID_CLASSES = [CoreSWHID, QualifiedSWHID, ExtendedSWHID]


@pytest.mark.parametrize(
    "invalid_swhid,swhid_class", itertools.product(INVALID_SWHIDS, SWHID_CLASSES)
)
def test_swhid_parsing_error(invalid_swhid, swhid_class):
    """Tests SWHID strings that are invalid for all SWHID classes do raise
    a ValidationError"""
    with pytest.raises(ValidationError):
        swhid_class.from_string(invalid_swhid)


# string SWHIDs, and how they should be parsed by each of the classes,
# or None if the class does not support it
HASH = "94a9ed024d3859793618152ea559a168bbcbb5e2"
VALID_SWHIDS = [
    (
        f"swh:1:cnt:{HASH}",
        CoreSWHID(
            object_type=ObjectType.CONTENT,
            object_id=_x(HASH),
        ),
        QualifiedSWHID(
            object_type=ObjectType.CONTENT,
            object_id=_x(HASH),
        ),
        ExtendedSWHID(
            object_type=ExtendedObjectType.CONTENT,
            object_id=_x(HASH),
        ),
    ),
    (
        f"swh:1:dir:{HASH}",
        CoreSWHID(
            object_type=ObjectType.DIRECTORY,
            object_id=_x(HASH),
        ),
        QualifiedSWHID(
            object_type=ObjectType.DIRECTORY,
            object_id=_x(HASH),
        ),
        ExtendedSWHID(
            object_type=ExtendedObjectType.DIRECTORY,
            object_id=_x(HASH),
        ),
    ),
    (
        f"swh:1:rev:{HASH}",
        CoreSWHID(
            object_type=ObjectType.REVISION,
            object_id=_x(HASH),
        ),
        QualifiedSWHID(
            object_type=ObjectType.REVISION,
            object_id=_x(HASH),
        ),
        ExtendedSWHID(
            object_type=ExtendedObjectType.REVISION,
            object_id=_x(HASH),
        ),
    ),
    (
        f"swh:1:rel:{HASH}",
        CoreSWHID(
            object_type=ObjectType.RELEASE,
            object_id=_x(HASH),
        ),
        QualifiedSWHID(
            object_type=ObjectType.RELEASE,
            object_id=_x(HASH),
        ),
        ExtendedSWHID(
            object_type=ExtendedObjectType.RELEASE,
            object_id=_x(HASH),
        ),
    ),
    (
        f"swh:1:snp:{HASH}",
        CoreSWHID(
            object_type=ObjectType.SNAPSHOT,
            object_id=_x(HASH),
        ),
        QualifiedSWHID(
            object_type=ObjectType.SNAPSHOT,
            object_id=_x(HASH),
        ),
        ExtendedSWHID(
            object_type=ExtendedObjectType.SNAPSHOT,
            object_id=_x(HASH),
        ),
    ),
    (
        f"swh:1:cnt:{HASH};origin=https://github.com/python/cpython;lines=1-18",
        None,  # CoreSWHID does not allow qualifiers
        QualifiedSWHID(
            object_type=ObjectType.CONTENT,
            object_id=_x(HASH),
            origin="https://github.com/python/cpython",
            lines=(1, 18),
        ),
        None,  # Neither does ExtendedSWHID
    ),
    (
        f"swh:1:cnt:{HASH};origin=https://github.com/python/cpython;lines=1-18/",
        None,  # likewise
        None,
        None,  # likewise
    ),
    (
        f"swh:1:cnt:{HASH};origin=https://github.com/python/cpython;lines=18",
        None,  # likewise
        QualifiedSWHID(
            object_type=ObjectType.CONTENT,
            object_id=_x(HASH),
            origin="https://github.com/python/cpython",
            lines=(18, None),
        ),
        None,  # likewise
    ),
    (
        f"swh:1:dir:{HASH};origin=deb://Debian/packages/linuxdoc-tools",
        None,  # likewise
        QualifiedSWHID(
            object_type=ObjectType.DIRECTORY,
            object_id=_x(HASH),
            origin="deb://Debian/packages/linuxdoc-tools",
        ),
        None,  # likewise
    ),
    (
        f"swh:1:ori:{HASH}",
        None,  # CoreSWHID does not allow origin pseudo-SWHIDs
        None,  # Neither does QualifiedSWHID
        ExtendedSWHID(
            object_type=ExtendedObjectType.ORIGIN,
            object_id=_x(HASH),
        ),
    ),
    (
        f"swh:1:emd:{HASH}",
        None,  # likewise for metadata pseudo-SWHIDs
        None,  # Neither does QualifiedSWHID
        ExtendedSWHID(
            object_type=ExtendedObjectType.RAW_EXTRINSIC_METADATA,
            object_id=_x(HASH),
        ),
    ),
    (
        f"swh:1:emd:{HASH};origin=https://github.com/python/cpython",
        None,  # CoreSWHID does not allow metadata pseudo-SWHIDs or qualifiers
        None,  # QualifiedSWHID does not allow metadata pseudo-SWHIDs
        None,  # ExtendedSWHID does not allow qualifiers
    ),
]


@pytest.mark.parametrize(
    "string,core,qualified,extended",
    [
        pytest.param(string, core, qualified, extended, id=string)
        for (string, core, qualified, extended) in VALID_SWHIDS
    ],
)
def test_parse_unparse_swhids(string, core, qualified, extended):
    """Tests parsing and serializing valid SWHIDs with the various SWHID classes."""
    classes = [CoreSWHID, QualifiedSWHID, ExtendedSWHID]
    for (cls, parsed_swhid) in zip(classes, [core, qualified, extended]):
        if parsed_swhid is None:
            # This class should not accept this SWHID
            with pytest.raises(ValidationError) as excinfo:
                cls.from_string(string)
            # Check string serialization for exception
            assert str(excinfo.value) is not None
        else:
            # This class should
            assert cls.from_string(string) == parsed_swhid

            # Also check serialization
            assert string == str(parsed_swhid)


@pytest.mark.parametrize(
    "core,extended",
    [
        pytest.param(core, extended, id=string)
        for (string, core, qualified, extended) in VALID_SWHIDS
        if core is not None
    ],
)
def test_core_to_extended(core, extended):
    assert core.to_extended() == extended


@pytest.mark.parametrize(
    "ns,version,type,id,qualifiers",
    [
        ("foo", 1, ObjectType.CONTENT, "abc8bc9d7a6bcf6db04f476d29314f157507d505", {}),
        ("swh", 2, ObjectType.CONTENT, "def8bc9d7a6bcf6db04f476d29314f157507d505", {}),
        ("swh", 1, ObjectType.DIRECTORY, "aaaa", {}),
    ],
)
def test_QualifiedSWHID_validation_error(ns, version, type, id, qualifiers):
    with pytest.raises(ValidationError):
        QualifiedSWHID(
            namespace=ns,
            scheme_version=version,
            object_type=type,
            object_id=_x(id),
            **qualifiers,
        )


@pytest.mark.parametrize(
    "object_type,qualifiers,expected",
    [
        # No qualifier:
        (ObjectType.CONTENT, {}, f"swh:1:cnt:{HASH}"),
        # origin:
        (ObjectType.CONTENT, {"origin": None}, f"swh:1:cnt:{HASH}"),
        (ObjectType.CONTENT, {"origin": 42}, ValueError),
        # visit:
        (
            ObjectType.CONTENT,
            {"visit": f"swh:1:snp:{HASH}"},
            f"swh:1:cnt:{HASH};visit=swh:1:snp:{HASH}",
        ),
        (
            ObjectType.CONTENT,
            {"visit": CoreSWHID(object_type=ObjectType.SNAPSHOT, object_id=_x(HASH))},
            f"swh:1:cnt:{HASH};visit=swh:1:snp:{HASH}",
        ),
        (ObjectType.CONTENT, {"visit": 42}, TypeError),
        (
            ObjectType.CONTENT,
            {"visit": f"swh:1:rel:{HASH}"},
            ValidationError,
        ),
        (
            ObjectType.CONTENT,
            {"visit": CoreSWHID(object_type=ObjectType.RELEASE, object_id=_x(HASH))},
            ValidationError,
        ),
        # anchor:
        (
            ObjectType.CONTENT,
            {"anchor": f"swh:1:snp:{HASH}"},
            f"swh:1:cnt:{HASH};anchor=swh:1:snp:{HASH}",
        ),
        (
            ObjectType.CONTENT,
            {"anchor": CoreSWHID(object_type=ObjectType.SNAPSHOT, object_id=_x(HASH))},
            f"swh:1:cnt:{HASH};anchor=swh:1:snp:{HASH}",
        ),
        (
            ObjectType.CONTENT,
            {"anchor": f"swh:1:dir:{HASH}"},
            f"swh:1:cnt:{HASH};anchor=swh:1:dir:{HASH}",
        ),
        (
            ObjectType.CONTENT,
            {"anchor": CoreSWHID(object_type=ObjectType.DIRECTORY, object_id=_x(HASH))},
            f"swh:1:cnt:{HASH};anchor=swh:1:dir:{HASH}",
        ),
        (ObjectType.CONTENT, {"anchor": 42}, TypeError),
        (
            ObjectType.CONTENT,
            {"anchor": f"swh:1:cnt:{HASH}"},
            ValidationError,
        ),
        (
            ObjectType.CONTENT,
            {"anchor": CoreSWHID(object_type=ObjectType.CONTENT, object_id=_x(HASH))},
            ValidationError,
        ),
        # path:
        (
            ObjectType.CONTENT,
            {"path": b"/foo"},
            f"swh:1:cnt:{HASH};path=/foo",
        ),
        (
            ObjectType.CONTENT,
            {"path": b"/foo;bar"},
            f"swh:1:cnt:{HASH};path=/foo%3Bbar",
        ),
        (
            ObjectType.CONTENT,
            {"path": "/foo"},
            f"swh:1:cnt:{HASH};path=/foo",
        ),
        (
            ObjectType.CONTENT,
            {"path": "/foo;bar"},
            f"swh:1:cnt:{HASH};path=/foo%3Bbar",
        ),
        (ObjectType.CONTENT, {"path": 42}, Exception),
        # lines:
        (
            ObjectType.CONTENT,
            {"lines": (42, None)},
            f"swh:1:cnt:{HASH};lines=42",
        ),
        (
            ObjectType.CONTENT,
            {"lines": (21, 42)},
            f"swh:1:cnt:{HASH};lines=21-42",
        ),
        (
            ObjectType.CONTENT,
            {"lines": 42},
            TypeError,
        ),
        (
            ObjectType.CONTENT,
            {"lines": (None, 42)},
            ValueError,
        ),
        (
            ObjectType.CONTENT,
            {"lines": ("42", None)},
            ValueError,
        ),
    ],
)
def test_QualifiedSWHID_init(object_type, qualifiers, expected):
    """Tests validation and converters of qualifiers"""
    if isinstance(expected, type):
        assert issubclass(expected, Exception)
        with pytest.raises(expected):
            QualifiedSWHID(object_type=object_type, object_id=_x(HASH), **qualifiers)
    else:
        assert isinstance(expected, str)
        swhid = QualifiedSWHID(
            object_type=object_type, object_id=_x(HASH), **qualifiers
        )

        # Check the build object has the right serialization
        assert expected == str(swhid)

        # Check the internal state of the object is the same as if parsed from a string
        assert QualifiedSWHID.from_string(expected) == swhid


def test_QualifiedSWHID_hash():
    object_id = _x("94a9ed024d3859793618152ea559a168bbcbb5e2")

    assert hash(
        QualifiedSWHID(object_type=ObjectType.DIRECTORY, object_id=object_id)
    ) == hash(QualifiedSWHID(object_type=ObjectType.DIRECTORY, object_id=object_id))

    assert hash(
        QualifiedSWHID(
            object_type=ObjectType.DIRECTORY,
            object_id=object_id,
            **dummy_qualifiers,
        )
    ) == hash(
        QualifiedSWHID(
            object_type=ObjectType.DIRECTORY,
            object_id=object_id,
            **dummy_qualifiers,
        )
    )

    # Different order of the dictionary, so the underlying order of the tuple in
    # ImmutableDict is different.
    assert hash(
        QualifiedSWHID(
            object_type=ObjectType.DIRECTORY,
            object_id=object_id,
            origin="https://example.com",
            lines=(42, None),
        )
    ) == hash(
        QualifiedSWHID(
            object_type=ObjectType.DIRECTORY,
            object_id=object_id,
            lines=(42, None),
            origin="https://example.com",
        )
    )


def test_QualifiedSWHID_eq():
    object_id = _x("94a9ed024d3859793618152ea559a168bbcbb5e2")

    assert QualifiedSWHID(
        object_type=ObjectType.DIRECTORY, object_id=object_id
    ) == QualifiedSWHID(object_type=ObjectType.DIRECTORY, object_id=object_id)

    assert QualifiedSWHID(
        object_type=ObjectType.DIRECTORY,
        object_id=object_id,
        **dummy_qualifiers,
    ) == QualifiedSWHID(
        object_type=ObjectType.DIRECTORY,
        object_id=object_id,
        **dummy_qualifiers,
    )

    assert QualifiedSWHID(
        object_type=ObjectType.DIRECTORY,
        object_id=object_id,
        **dummy_qualifiers,
    ) == QualifiedSWHID(
        object_type=ObjectType.DIRECTORY,
        object_id=object_id,
        **dummy_qualifiers,
    )


QUALIFIED_SWHIDS = [
    # origin:
    (
        f"swh:1:cnt:{HASH};origin=https://github.com/python/cpython",
        QualifiedSWHID(
            object_type=ObjectType.CONTENT,
            object_id=_x(HASH),
            origin="https://github.com/python/cpython",
        ),
    ),
    (
        f"swh:1:cnt:{HASH};origin=https://example.org/foo%3Bbar%25baz",
        QualifiedSWHID(
            object_type=ObjectType.CONTENT,
            object_id=_x(HASH),
            origin="https://example.org/foo%3Bbar%25baz",
        ),
    ),
    (
        f"swh:1:cnt:{HASH};origin=https://example.org?project=test",
        QualifiedSWHID(
            object_type=ObjectType.CONTENT,
            object_id=_x(HASH),
            origin="https://example.org?project=test",
        ),
    ),
    # visit:
    (
        f"swh:1:cnt:{HASH};visit=swh:1:snp:{HASH}",
        QualifiedSWHID(
            object_type=ObjectType.CONTENT,
            object_id=_x(HASH),
            visit=CoreSWHID(object_type=ObjectType.SNAPSHOT, object_id=_x(HASH)),
        ),
    ),
    (
        f"swh:1:cnt:{HASH};visit=swh:1:rel:{HASH}",
        None,
    ),
    # anchor:
    (
        f"swh:1:cnt:{HASH};anchor=swh:1:dir:{HASH}",
        QualifiedSWHID(
            object_type=ObjectType.CONTENT,
            object_id=_x(HASH),
            anchor=CoreSWHID(object_type=ObjectType.DIRECTORY, object_id=_x(HASH)),
        ),
    ),
    (
        f"swh:1:cnt:{HASH};anchor=swh:1:rev:{HASH}",
        QualifiedSWHID(
            object_type=ObjectType.CONTENT,
            object_id=_x(HASH),
            anchor=CoreSWHID(object_type=ObjectType.REVISION, object_id=_x(HASH)),
        ),
    ),
    (
        f"swh:1:cnt:{HASH};anchor=swh:1:cnt:{HASH}",
        None,  # 'cnt' is not valid in anchor
    ),
    (
        f"swh:1:cnt:{HASH};anchor=swh:1:ori:{HASH}",
        None,  # 'ori' is not valid in a CoreSWHID
    ),
    # path:
    (
        f"swh:1:cnt:{HASH};path=/foo",
        QualifiedSWHID(
            object_type=ObjectType.CONTENT, object_id=_x(HASH), path=b"/foo"
        ),
    ),
    (
        f"swh:1:cnt:{HASH};path=/foo%3Bbar",
        QualifiedSWHID(
            object_type=ObjectType.CONTENT, object_id=_x(HASH), path=b"/foo;bar"
        ),
    ),
    (
        f"swh:1:cnt:{HASH};path=/foo%25bar",
        QualifiedSWHID(
            object_type=ObjectType.CONTENT, object_id=_x(HASH), path=b"/foo%bar"
        ),
    ),
    (
        f"swh:1:cnt:{HASH};path=/foo/bar%3Dbaz",
        QualifiedSWHID(
            object_type=ObjectType.CONTENT, object_id=_x(HASH), path=b"/foo/bar=baz"
        ),
    ),
    # lines
    (
        f"swh:1:cnt:{HASH};lines=1-18",
        QualifiedSWHID(
            object_type=ObjectType.CONTENT,
            object_id=_x(HASH),
            lines=(1, 18),
        ),
    ),
    (
        f"swh:1:cnt:{HASH};lines=18",
        QualifiedSWHID(
            object_type=ObjectType.CONTENT,
            object_id=_x(HASH),
            lines=(18, None),
        ),
    ),
    (
        f"swh:1:cnt:{HASH};lines=",
        None,
    ),
    (
        f"swh:1:cnt:{HASH};lines=aa",
        None,
    ),
    (
        f"swh:1:cnt:{HASH};lines=18-aa",
        None,
    ),
]


@pytest.mark.parametrize("string,parsed", QUALIFIED_SWHIDS)
def test_QualifiedSWHID_parse_serialize_qualifiers(string, parsed):
    """Tests parsing and serializing valid SWHIDs with the various SWHID classes."""
    if parsed is None:
        with pytest.raises(ValidationError):
            print(repr(QualifiedSWHID.from_string(string)))
    else:
        assert QualifiedSWHID.from_string(string) == parsed
        assert str(parsed) == string


def test_QualifiedSWHID_serialize_origin():
    """Checks that semicolon in origins are escaped."""
    string = f"swh:1:cnt:{HASH};origin=https://example.org/foo%3Bbar%25baz"
    swhid = QualifiedSWHID(
        object_type=ObjectType.CONTENT,
        object_id=_x(HASH),
        origin="https://example.org/foo;bar%25baz",
    )
    assert str(swhid) == string


def test_QualifiedSWHID_attributes():
    """Checks the set of QualifiedSWHID attributes match the SWHID_QUALIFIERS
    constant."""

    assert set(attr.fields_dict(QualifiedSWHID)) == {
        "namespace",
        "scheme_version",
        "object_type",
        "object_id",
        *SWHID_QUALIFIERS,
    }


@pytest.mark.parametrize(
    "ns,version,type,id",
    [
        ("foo", 1, ObjectType.CONTENT, "abc8bc9d7a6bcf6db04f476d29314f157507d505"),
        ("swh", 2, ObjectType.CONTENT, "def8bc9d7a6bcf6db04f476d29314f157507d505"),
        ("swh", 1, ObjectType.DIRECTORY, "aaaa"),
    ],
)
def test_CoreSWHID_validation_error(ns, version, type, id):
    with pytest.raises(ValidationError):
        CoreSWHID(
            namespace=ns,
            scheme_version=version,
            object_type=type,
            object_id=_x(id),
        )


def test_CoreSWHID_hash():
    object_id = _x("94a9ed024d3859793618152ea559a168bbcbb5e2")

    assert hash(
        CoreSWHID(object_type=ObjectType.DIRECTORY, object_id=object_id)
    ) == hash(CoreSWHID(object_type=ObjectType.DIRECTORY, object_id=object_id))

    assert hash(
        CoreSWHID(
            object_type=ObjectType.DIRECTORY,
            object_id=object_id,
        )
    ) == hash(
        CoreSWHID(
            object_type=ObjectType.DIRECTORY,
            object_id=object_id,
        )
    )

    # Different order of the dictionary, so the underlying order of the tuple in
    # ImmutableDict is different.
    assert hash(
        CoreSWHID(
            object_type=ObjectType.DIRECTORY,
            object_id=object_id,
        )
    ) == hash(
        CoreSWHID(
            object_type=ObjectType.DIRECTORY,
            object_id=object_id,
        )
    )


def test_CoreSWHID_eq():
    object_id = _x("94a9ed024d3859793618152ea559a168bbcbb5e2")

    assert CoreSWHID(
        object_type=ObjectType.DIRECTORY, object_id=object_id
    ) == CoreSWHID(object_type=ObjectType.DIRECTORY, object_id=object_id)

    assert CoreSWHID(
        object_type=ObjectType.DIRECTORY,
        object_id=object_id,
    ) == CoreSWHID(
        object_type=ObjectType.DIRECTORY,
        object_id=object_id,
    )

    assert CoreSWHID(
        object_type=ObjectType.DIRECTORY,
        object_id=object_id,
    ) == CoreSWHID(
        object_type=ObjectType.DIRECTORY,
        object_id=object_id,
    )


@pytest.mark.parametrize(
    "ns,version,type,id",
    [
        (
            "foo",
            1,
            ExtendedObjectType.CONTENT,
            "abc8bc9d7a6bcf6db04f476d29314f157507d505",
        ),
        (
            "swh",
            2,
            ExtendedObjectType.CONTENT,
            "def8bc9d7a6bcf6db04f476d29314f157507d505",
        ),
        ("swh", 1, ExtendedObjectType.DIRECTORY, "aaaa"),
    ],
)
def test_ExtendedSWHID_validation_error(ns, version, type, id):
    with pytest.raises(ValidationError):
        ExtendedSWHID(
            namespace=ns,
            scheme_version=version,
            object_type=type,
            object_id=_x(id),
        )


def test_ExtendedSWHID_hash():
    object_id = _x("94a9ed024d3859793618152ea559a168bbcbb5e2")

    assert hash(
        ExtendedSWHID(object_type=ExtendedObjectType.DIRECTORY, object_id=object_id)
    ) == hash(
        ExtendedSWHID(object_type=ExtendedObjectType.DIRECTORY, object_id=object_id)
    )

    assert hash(
        ExtendedSWHID(
            object_type=ExtendedObjectType.DIRECTORY,
            object_id=object_id,
        )
    ) == hash(
        ExtendedSWHID(
            object_type=ExtendedObjectType.DIRECTORY,
            object_id=object_id,
        )
    )

    # Different order of the dictionary, so the underlying order of the tuple in
    # ImmutableDict is different.
    assert hash(
        ExtendedSWHID(
            object_type=ExtendedObjectType.DIRECTORY,
            object_id=object_id,
        )
    ) == hash(
        ExtendedSWHID(
            object_type=ExtendedObjectType.DIRECTORY,
            object_id=object_id,
        )
    )


def test_ExtendedSWHID_eq():
    object_id = _x("94a9ed024d3859793618152ea559a168bbcbb5e2")

    assert ExtendedSWHID(
        object_type=ExtendedObjectType.DIRECTORY, object_id=object_id
    ) == ExtendedSWHID(object_type=ExtendedObjectType.DIRECTORY, object_id=object_id)

    assert ExtendedSWHID(
        object_type=ExtendedObjectType.DIRECTORY,
        object_id=object_id,
    ) == ExtendedSWHID(
        object_type=ExtendedObjectType.DIRECTORY,
        object_id=object_id,
    )

    assert ExtendedSWHID(
        object_type=ExtendedObjectType.DIRECTORY,
        object_id=object_id,
    ) == ExtendedSWHID(
        object_type=ExtendedObjectType.DIRECTORY,
        object_id=object_id,
    )


def test_object_types():
    """Checks ExtendedObjectType is a superset of ObjectType"""
    for member in ObjectType:
        assert getattr(ExtendedObjectType, member.name).value == member.value



# Copyright (C) 2017-2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Merkle tree data structure"""

from __future__ import annotations

import abc
from typing import Any, Dict, Iterator, List, Set


class MerkleNode(dict, metaclass=abc.ABCMeta):
    """Representation of a node in a Merkle Tree.

    A (generalized) `Merkle Tree`_ is a tree in which every node is labeled
    with a hash of its own data and the hash of its children.

    .. _Merkle Tree: https://en.wikipedia.org/wiki/Merkle_tree

    In pseudocode::

      node.hash = hash(node.data
                       + sum(child.hash for child in node.children))

    This class efficiently implements the Merkle Tree data structure on top of
    a Python :class:`dict`, minimizing hash computations and new data
    collections when updating nodes.

    Node data is stored in the :attr:`data` attribute, while (named) children
    are stored as items of the underlying dictionary.

    Addition, update and removal of objects are instrumented to automatically
    invalidate the hashes of the current node as well as its registered
    parents; It also resets the collection status of the objects so the updated
    objects can be collected.

    The collection of updated data from the tree is implemented through the
    :func:`collect` function and associated helpers.

    """

    __slots__ = ["parents", "data", "__hash", "collected"]

    data: Dict
    """data associated to the current node"""

    parents: List
    """known parents of the current node"""

    collected: bool
    """whether the current node has been collected"""

    def __init__(self, data=None):
        super().__init__()
        self.parents = []
        self.data = data
        self.__hash = None
        self.collected = False

    def __eq__(self, other):
        return (
            isinstance(other, MerkleNode)
            and super().__eq__(other)
            and self.data == other.data
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def invalidate_hash(self):
        """Invalidate the cached hash of the current node."""
        if not self.__hash:
            return

        self.__hash = None
        self.collected = False
        for parent in self.parents:
            parent.invalidate_hash()

    def update_hash(self, *, force=False) -> Any:
        """Recursively compute the hash of the current node.

        Args:
          force (bool): invalidate the cache and force the computation for
            this node and all children.
        """
        if self.__hash and not force:
            return self.__hash

        if force:
            self.invalidate_hash()

        for child in self.values():
            child.update_hash(force=force)

        self.__hash = self.compute_hash()
        return self.__hash

    @property
    def hash(self) -> Any:
        """The hash of the current node, as calculated by
        :func:`compute_hash`.
        """
        return self.update_hash()

    def __hash__(self):
        return hash(self.hash)

    @abc.abstractmethod
    def compute_hash(self) -> Any:
        """Compute the hash of the current node.

        The hash should depend on the data of the node, as well as on hashes
        of the children nodes.
        """
        raise NotImplementedError("Must implement compute_hash method")

    def __setitem__(self, name, new_child):
        """Add a child, invalidating the current hash"""
        self.invalidate_hash()

        super().__setitem__(name, new_child)

        new_child.parents.append(self)

    def __delitem__(self, name):
        """Remove a child, invalidating the current hash"""
        if name in self:
            self.invalidate_hash()
            self[name].parents.remove(self)
            super().__delitem__(name)
        else:
            raise KeyError(name)

    def update(self, new_children):
        """Add several named children from a dictionary"""
        if not new_children:
            return

        self.invalidate_hash()

        for name, new_child in new_children.items():
            new_child.parents.append(self)
            if name in self:
                self[name].parents.remove(self)

        super().update(new_children)

    def get_data(self, **kwargs):
        """Retrieve and format the collected data for the current node, for use by
        :func:`collect`.

        Can be overridden, for instance when you want the collected data to
        contain information about the child nodes.

        Arguments:
          kwargs: allow subclasses to alter behaviour depending on how
            :func:`collect` is called.

        Returns:
          data formatted for :func:`collect`
        """
        return self.data

    def collect_node(self) -> Set[MerkleNode]:
        """Collect the current node if it has not been yet, for use by :func:`collect`."""
        if not self.collected:
            self.collected = True
            return {self}
        else:
            return set()

    def collect(self) -> Set[MerkleNode]:
        """Collect the added and modified nodes in the subtree rooted at `self`
        since the last collect operation.

        Returns:
           A :class:`set` of collected nodes
        """
        ret = self.collect_node()
        for child in self.values():
            ret.update(child.collect())

        return ret

    def reset_collect(self):
        """Recursively unmark collected nodes in the subtree rooted at `self`.

        This lets the caller use :func:`collect` again.
        """
        self.collected = False

        for child in self.values():
            child.reset_collect()

    def iter_tree(self, dedup=True) -> Iterator[MerkleNode]:
        """Yields all children nodes, recursively. Common nodes are deduplicated
        by default (deduplication can be turned off setting the given argument
        'dedup' to False).
        """
        yield from self._iter_tree(set(), dedup)

    def _iter_tree(self, seen: Set[bytes], dedup) -> Iterator[MerkleNode]:
        if self.hash not in seen:
            if dedup:
                seen.add(self.hash)
            yield self
            for child in self.values():
                yield from child._iter_tree(seen=seen, dedup=dedup)


class MerkleLeaf(MerkleNode):
    """A leaf to a Merkle tree.

    A Merkle leaf is simply a Merkle node with children disabled.
    """

    __slots__ = []  # type: List[str]

    def __setitem__(self, name, child):
        raise ValueError("%s is a leaf" % self.__class__.__name__)

    def __getitem__(self, name):
        raise ValueError("%s is a leaf" % self.__class__.__name__)

    def __delitem__(self, name):
        raise ValueError("%s is a leaf" % self.__class__.__name__)

    def update(self, new_children):
        """Children update operation. Disabled for leaves."""
        raise ValueError("%s is a leaf" % self.__class__.__name__)



# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import numbers

from ..exceptions import ValidationError


def validate_type(value, type):
    """Validate that value is an integer"""
    if not isinstance(value, type):
        if isinstance(type, tuple):
            typestr = "one of %s" % ", ".join(typ.__name__ for typ in type)
        else:
            typestr = type.__name__
        raise ValidationError(
            "Unexpected type %(type)s, expected %(expected_type)s",
            params={
                "type": value.__class__.__name__,
                "expected_type": typestr,
            },
            code="unexpected-type",
        )

    return True


def validate_int(value):
    """Validate that the given value is an int"""
    return validate_type(value, numbers.Integral)


def validate_str(value):
    """Validate that the given value is a string"""
    return validate_type(value, str)


def validate_bytes(value):
    """Validate that the given value is a bytes object"""
    return validate_type(value, bytes)


def validate_datetime(value):
    """Validate that the given value is either a datetime, or a numeric number
    of seconds since the UNIX epoch."""

    errors = []
    try:
        validate_type(value, (datetime.datetime, numbers.Real))
    except ValidationError as e:
        errors.append(e)

    if isinstance(value, datetime.datetime) and value.tzinfo is None:
        errors.append(
            ValidationError(
                "Datetimes must be timezone-aware in swh",
                code="datetime-without-tzinfo",
            )
        )

    if errors:
        raise ValidationError(errors)

    return True


def validate_enum(value, expected_values):
    """Validate that value is contained in expected_values"""
    if value not in expected_values:
        raise ValidationError(
            "Unexpected value %(value)s, expected one of %(expected_values)s",
            params={
                "value": value,
                "expected_values": ", ".join(sorted(expected_values)),
            },
            code="unexpected-value",
        )

    return True



# Copyright (C) 2015-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Dict
import warnings

from . import model

# Reexport for backward compatibility
from .git_objects import *  # noqa
from .hashutil import MultiHash, hash_to_hex

# Reexport for backward compatibility
from .swhids import *  # noqa

warnings.warn(
    "The swh.model.identifiers module is deprecated. "
    "SWHID-related classes were moved to swh.model.swhids, and identifier "
    "computation is now done directly with swh.model.model classes.",
    DeprecationWarning,
    stacklevel=2,
)

# The following are deprecated aliases of the variants defined in ObjectType
# while transitioning from SWHID to QualifiedSWHID
ORIGIN = "origin"
SNAPSHOT = "snapshot"
REVISION = "revision"
RELEASE = "release"
DIRECTORY = "directory"
CONTENT = "content"
RAW_EXTRINSIC_METADATA = "raw_extrinsic_metadata"


def content_identifier(content: Dict[str, Any]) -> Dict[str, bytes]:
    """Deprecated, use :class:`swh.model.Content` instead:
    ``content_identifier(d)`` is equivalent to:
    ``{k: hash_to_hex(v) for (k, v) in Content.from_data(d["data"]).hashes().items()}``
    """
    return MultiHash.from_data(content["data"]).digest()


def directory_identifier(directory: Dict[str, Any]) -> str:
    """Deprecated, use :class:`swh.model.Directory` instead:
    ``directory_identifier(d)`` is equivalent to:
    ``hash_to_hex(Directory.from_dict(d).id)``.

    See :func:`swh.model.git_objects.directory_git_object` for details of the
    format used to generate this identifier."""
    return hash_to_hex(model.Directory.from_dict(directory).id)


def revision_identifier(revision: Dict[str, Any]) -> str:
    """Deprecated, use :class:`swh.model.Revision` instead:
    ``revision_identifier(d)`` is equivalent to:
    ``hash_to_hex(Revision.from_dict(d).id)``.

    See :func:`swh.model.git_objects.revision_git_object` for details of the
    format used to generate this identifier."""
    return hash_to_hex(model.Revision.from_dict(revision).id)


def release_identifier(release: Dict[str, Any]) -> str:
    """Deprecated, use :class:`swh.model.Release` instead:
    ``release_identifier(d)`` is equivalent to:
    ``hash_to_hex(Release.from_dict(d).id)``.

    See :func:`swh.model.git_objects.release_git_object` for details of the
    format used to generate this identifier."""
    return hash_to_hex(model.Release.from_dict(release).id)


def snapshot_identifier(snapshot: Dict[str, Any]) -> str:
    """Deprecated, use :class:`swh.model.Snapshot` instead:
    ``snapshot_identifier(d)`` is equivalent to:
    ``hash_to_hex(Snapshot.from_dict(d).id)``.

    See :func:`swh.model.git_objects.snapshot_git_object` for details of the
    format used to generate this identifier."""
    return hash_to_hex(model.Snapshot.from_dict(snapshot).id)


def origin_identifier(origin):
    """Deprecated, use :class:`swh.model.Origin` instead:
    ``origin_identifier(url)`` is equivalent to:
    ``hash_to_hex(Origin(url=url).id)``.
    """

    return hash_to_hex(model.Origin.from_dict(origin).id)



# Copyright (C) 2017-2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest

from swh.model import merkle


class MerkleTestNode(merkle.MerkleNode):
    object_type = "tested_merkle_node_type"

    def __init__(self, data):
        super().__init__(data)
        self.compute_hash_called = 0

    def compute_hash(self) -> bytes:
        self.compute_hash_called += 1
        child_data = [child + b"=" + self[child].hash for child in sorted(self)]
        return b"hash(" + b", ".join([self.data.get("value", b"")] + child_data) + b")"


class MerkleTestLeaf(merkle.MerkleLeaf):
    object_type = "tested_merkle_leaf_type"

    def __init__(self, data):
        super().__init__(data)
        self.compute_hash_called = 0

    def compute_hash(self):
        self.compute_hash_called += 1
        return b"hash(" + self.data.get("value", b"") + b")"


class TestMerkleLeaf(unittest.TestCase):
    def setUp(self):
        self.data = {"value": b"value"}
        self.instance = MerkleTestLeaf(self.data)

    def test_equality(self):
        leaf1 = MerkleTestLeaf(self.data)
        leaf2 = MerkleTestLeaf(self.data)
        leaf3 = MerkleTestLeaf({})

        self.assertEqual(leaf1, leaf2)
        self.assertNotEqual(leaf1, leaf3)

    def test_hash(self):
        self.assertEqual(self.instance.compute_hash_called, 0)
        instance_hash = self.instance.hash
        self.assertEqual(self.instance.compute_hash_called, 1)
        instance_hash2 = self.instance.hash
        self.assertEqual(self.instance.compute_hash_called, 1)
        self.assertEqual(instance_hash, instance_hash2)

    def test_data(self):
        self.assertEqual(self.instance.get_data(), self.data)

    def test_collect(self):
        collected = self.instance.collect()
        self.assertEqual(
            collected,
            {self.instance},
        )
        collected2 = self.instance.collect()
        self.assertEqual(collected2, set())
        self.instance.reset_collect()
        collected3 = self.instance.collect()
        self.assertEqual(collected, collected3)

    def test_leaf(self):
        with self.assertRaisesRegex(ValueError, "is a leaf"):
            self.instance[b"key1"] = "Test"

        with self.assertRaisesRegex(ValueError, "is a leaf"):
            del self.instance[b"key1"]

        with self.assertRaisesRegex(ValueError, "is a leaf"):
            self.instance[b"key1"]

        with self.assertRaisesRegex(ValueError, "is a leaf"):
            self.instance.update(self.data)


class TestMerkleNode(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        self.root = MerkleTestNode({"value": b"root"})
        self.nodes = {b"root": self.root}
        for i in (b"a", b"b", b"c"):
            value = b"root/" + i
            node = MerkleTestNode(
                {
                    "value": value,
                }
            )
            self.root[i] = node
            self.nodes[value] = node
            for j in (b"a", b"b", b"c"):
                value2 = value + b"/" + j
                node2 = MerkleTestNode(
                    {
                        "value": value2,
                    }
                )
                node[j] = node2
                self.nodes[value2] = node2
                for k in (b"a", b"b", b"c"):
                    value3 = value2 + b"/" + j
                    node3 = MerkleTestNode(
                        {
                            "value": value3,
                        }
                    )
                    node2[j] = node3
                    self.nodes[value3] = node3

    def test_equality(self):
        node1 = MerkleTestNode({"value": b"bar"})
        node2 = MerkleTestNode({"value": b"bar"})
        node3 = MerkleTestNode({})

        self.assertEqual(node1, node2)
        self.assertNotEqual(node1, node3, node1 == node3)

        node1[b"a"] = node3
        self.assertNotEqual(node1, node2)

        node2[b"a"] = node3
        self.assertEqual(node1, node2)

    def test_hash(self):
        for node in self.nodes.values():
            self.assertEqual(node.compute_hash_called, 0)

        # Root hash will compute hash for all the nodes
        hash = self.root.hash
        for node in self.nodes.values():
            self.assertEqual(node.compute_hash_called, 1)
            self.assertIn(node.data["value"], hash)

        # Should use the cached value
        hash2 = self.root.hash
        self.assertEqual(hash, hash2)
        for node in self.nodes.values():
            self.assertEqual(node.compute_hash_called, 1)

        # Should still use the cached value
        hash3 = self.root.update_hash(force=False)
        self.assertEqual(hash, hash3)
        for node in self.nodes.values():
            self.assertEqual(node.compute_hash_called, 1)

        # Force update of the cached value for a deeply nested node
        self.root[b"a"][b"b"].update_hash(force=True)
        for key, node in self.nodes.items():
            # update_hash rehashes all children
            if key.startswith(b"root/a/b"):
                self.assertEqual(node.compute_hash_called, 2)
            else:
                self.assertEqual(node.compute_hash_called, 1)

        hash4 = self.root.hash
        self.assertEqual(hash, hash4)
        for key, node in self.nodes.items():
            # update_hash also invalidates all parents
            if key in (b"root", b"root/a") or key.startswith(b"root/a/b"):
                self.assertEqual(node.compute_hash_called, 2)
            else:
                self.assertEqual(node.compute_hash_called, 1)

    def test_collect(self):
        collected = self.root.collect()
        self.assertEqual(collected, set(self.nodes.values()))
        for node in self.nodes.values():
            self.assertTrue(node.collected)
        collected2 = self.root.collect()
        self.assertEqual(collected2, set())

    def test_iter_tree_with_deduplication(self):
        nodes = list(self.root.iter_tree())
        self.assertCountEqual(nodes, self.nodes.values())

    def test_iter_tree_without_deduplication(self):
        # duplicate existing hash in merkle tree
        self.root[b"d"] = MerkleTestNode({"value": b"root/c/c/c"})
        nodes_dedup = list(self.root.iter_tree())
        nodes = list(self.root.iter_tree(dedup=False))
        assert nodes != nodes_dedup
        assert len(nodes) == len(nodes_dedup) + 1

    def test_get(self):
        for key in (b"a", b"b", b"c"):
            self.assertEqual(self.root[key], self.nodes[b"root/" + key])

        with self.assertRaisesRegex(KeyError, "b'nonexistent'"):
            self.root[b"nonexistent"]

    def test_del(self):
        hash_root = self.root.hash
        hash_a = self.nodes[b"root/a"].hash
        del self.root[b"a"][b"c"]
        hash_root2 = self.root.hash
        hash_a2 = self.nodes[b"root/a"].hash

        self.assertNotEqual(hash_root, hash_root2)
        self.assertNotEqual(hash_a, hash_a2)

        self.assertEqual(self.nodes[b"root/a/c"].parents, [])

        with self.assertRaisesRegex(KeyError, "b'nonexistent'"):
            del self.root[b"nonexistent"]

    def test_update(self):
        hash_root = self.root.hash
        hash_b = self.root[b"b"].hash
        new_children = {
            b"c": MerkleTestNode({"value": b"root/b/new_c"}),
            b"d": MerkleTestNode({"value": b"root/b/d"}),
        }

        # collect all nodes
        self.root.collect()

        self.root[b"b"].update(new_children)

        # Ensure everyone got reparented
        self.assertEqual(new_children[b"c"].parents, [self.root[b"b"]])
        self.assertEqual(new_children[b"d"].parents, [self.root[b"b"]])
        self.assertEqual(self.nodes[b"root/b/c"].parents, [])

        hash_root2 = self.root.hash
        self.assertNotEqual(hash_root, hash_root2)
        self.assertIn(b"root/b/new_c", hash_root2)
        self.assertIn(b"root/b/d", hash_root2)

        hash_b2 = self.root[b"b"].hash
        self.assertNotEqual(hash_b, hash_b2)

        for key, node in self.nodes.items():
            if key in (b"root", b"root/b"):
                self.assertEqual(node.compute_hash_called, 2)
            else:
                self.assertEqual(node.compute_hash_called, 1)

        # Ensure we collected root, root/b, and both new children
        collected_after_update = self.root.collect()
        self.assertEqual(
            collected_after_update,
            {
                self.nodes[b"root"],
                self.nodes[b"root/b"],
                new_children[b"c"],
                new_children[b"d"],
            },
        )

        # test that noop updates doesn't invalidate anything
        self.root[b"a"][b"b"].update({})
        self.assertEqual(self.root.collect(), set())



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import attr
import pytest

from swh.model.tests.swh_model_data import TEST_OBJECTS


@pytest.mark.parametrize("object_type, objects", TEST_OBJECTS.items())
def test_swh_model_data(object_type, objects):
    """checks model objects in swh_model_data are in correct shape"""
    assert objects
    for obj in objects:
        assert obj.object_type == object_type
        attr.validate(obj)


@pytest.mark.parametrize(
    "object_type",
    ("directory", "revision", "release", "snapshot"),
)
def test_swh_model_data_hash(object_type):
    for obj in TEST_OBJECTS[object_type]:
        assert (
            obj.compute_hash() == obj.id
        ), f"{obj.compute_hash().hex()} != {obj.id.hex()}"


def test_ensure_visit_status_date_consistency():
    """ensure origin-visit-status dates are more recent than their visit counterpart

    The origin-visit-status dates needs to be shifted slightly in the future from their
    visit dates counterpart. Otherwise, we are hitting storage-wise the "on conflict"
    ignore policy (because origin-visit-add creates an origin-visit-status with the same
    parameters from the origin-visit {origin, visit, date}...

    """
    visits = TEST_OBJECTS["origin_visit"]
    visit_statuses = TEST_OBJECTS["origin_visit_status"]
    for visit, visit_status in zip(visits, visit_statuses):
        assert visit.origin == visit_status.origin
        assert visit.visit == visit_status.visit
        assert visit.date < visit_status.date


def test_ensure_visit_status_snapshot_consistency():
    """ensure origin-visit-status snapshots exist in the test dataset"""
    snapshots = [snp.id for snp in TEST_OBJECTS["snapshot"]]
    for visit_status in TEST_OBJECTS["origin_visit_status"]:
        if visit_status.snapshot:
            assert visit_status.snapshot in snapshots



# Copyright (C) 2020-2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

"""Utility data structures."""

from collections.abc import Mapping
import copy
from typing import Dict, Generic, Iterable, Optional, Tuple, TypeVar, Union

KT = TypeVar("KT")
VT = TypeVar("VT")


class ImmutableDict(Mapping, Generic[KT, VT]):
    """A frozen dictionary.

    This class behaves like a dictionary, but internally stores objects in a tuple,
    so it is both immutable and hashable."""

    _data: Dict[KT, VT]

    def __init__(
        self,
        data: Union[Iterable[Tuple[KT, VT]], ImmutableDict[KT, VT], Dict[KT, VT]] = {},
    ):
        if isinstance(data, dict):
            self._data = data
        elif isinstance(data, ImmutableDict):
            self._data = data._data
        else:
            self._data = {k: v for k, v in data}

    @property
    def data(self):
        return tuple(self._data.items())

    def __repr__(self):
        return f"ImmutableDict({dict(self.data)!r})"

    def __getitem__(self, key):
        return self._data[key]

    def __iter__(self):
        for (k, v) in self.data:
            yield k

    def __len__(self):
        return len(self._data)

    def items(self):
        yield from self.data

    def __hash__(self):
        return hash(tuple(sorted(self.data)))

    def copy_pop(self, popped_key) -> Tuple[Optional[VT], ImmutableDict[KT, VT]]:
        """Returns a copy of this ImmutableDict without the given key,
        as well as the value associated to the key."""
        new_items = copy.deepcopy(self._data)
        popped_value = new_items.pop(popped_key, None)  # type: ignore
        return (popped_value, ImmutableDict(new_items))



# Copyright (C) 2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from dataclasses import dataclass
from typing import Iterable, List

from swh.model import discovery, model
from swh.model.hashutil import hash_to_bytes
from swh.model.model import Sha1Git
from swh.model.tests.test_identifiers import directory_example

pytest_plugins = ["aiohttp.pytest_plugin"]

UNKNOWN_HASH = hash_to_bytes("17140cb6109f1e3296dc52e2b2cd29bcb40e86be")
KNOWN_CONTENT_HASH = hash_to_bytes("e8e4106de42e2d5d5efab6a9422b9a8677c993c8")
KNOWN_DIRECTORY_HASH = hash_to_bytes("d7ed3d2c31d608823be58b1cbe57605310615231")
KNOWN_DIRECTORY_HASH_2 = hash_to_bytes("c76724e9a0be4b60f4bf0cb48b261df8eda94b1d")


@dataclass
class FakeArchive:
    contents: List[model.Content]
    skipped_contents: List[model.SkippedContent]
    directories: List[model.Directory]

    async def content_missing(self, contents: List[Sha1Git]) -> Iterable[Sha1Git]:
        return []

    async def skipped_content_missing(
        self, skipped_contents: List[Sha1Git]
    ) -> Iterable[Sha1Git]:
        """List skipped content missing from the archive by sha1"""
        return []

    async def directory_missing(self, directories: List[Sha1Git]) -> Iterable[Sha1Git]:
        """List directories missing from the archive by sha1"""
        return []


async def test_filter_known_objects(monkeypatch):
    # Test with smaller sample sizes to actually trigger the random sampling
    monkeypatch.setattr(discovery, "SAMPLE_SIZE", 1)

    base_directory = model.Directory.from_dict(directory_example)

    # Hardcoding another hash is enough since it's all that's being checked
    directory_data = directory_example.copy()
    directory_data["id"] = KNOWN_DIRECTORY_HASH_2
    other_directory = model.Directory.from_dict(directory_data)
    archive = FakeArchive(
        contents=[model.Content.from_data(b"blabla")],
        skipped_contents=[model.SkippedContent.from_data(b"blabla2", reason="reason")],
        directories=[
            base_directory,
            other_directory,
        ],
    )
    assert archive.contents[0].sha1_git == KNOWN_CONTENT_HASH
    assert archive.directories[0].id == KNOWN_DIRECTORY_HASH
    assert archive.directories[1].id == KNOWN_DIRECTORY_HASH_2
    (contents, skipped_contents, directories) = await discovery.filter_known_objects(
        archive
    )
    assert len(contents) == 0
    assert len(skipped_contents) == 0
    assert len(directories) == 0
